#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr  4 13:41:18 2021
This script plots SPI maps for GPCC 0.25 degree data over South-Western
Southern Africa.
To use this script, please ensure that maps.py and analysis.py are in a
subdirectory of the directory from which this script is run, named "gridfn"
All SPI files should be in the working directory, or a path to them should
be specified in the `set up` cell
For figures to be saved, please also edit the maps.py file to
@author: stefaan
"""
# %% imports

import re
import xarray as xr
import numpy as np
import seaborn as sns
import gridfn.maps as gm
import gridfn.analysis as gan

sns.set_context("notebook")
sns.set_style("darkgrid")

# %% constants

# default colours to use for discrete variable mapping

# orange red rose purple wine indigo blue cyan teal green olive sand black
# pale_grey

my15 = ['#EE7733', '#CC3311', '#CC6677', '#AA4499', '#882255', '#332288',
        '#0077BB', '#88CCEE', '#44AA99', '#117733', '#999933', '#DDCC77',
        '#000000', '#DDDDDD']


# %% set-up

# name of SPI file without extension or timescale
SPI = "../data/GPCC_V2020_025deg_SWSA_SPI"
ns = [3, 12, 24, 36, 60]   # time scales

SWSA = xr.open_mfdataset(SPI + "*.nc", concat_dim='timescale')

# %% remove pnts that are constant throughout the time series (sd = 0)

SWSA['SPI'] = SWSA.SPI.where(SWSA.SPI.var(dim='time') > 0)

# %% further analysis

to_decade = np.vectorize(lambda x: re.sub('[0-9].0$', '0s', x))

minyrds = gan.xr_np_time2yr(gan.time_min(SWSA.SPI)).to_dataset()
minyrds['decade'] = minyrds.time.astype('str')
# hack; function to use only works on np
minyrds['decade'].values = to_decade(minyrds.decade)

# %% plot for selected times

gm.cts_gridplot(SWSA.SPI.isel(time=-1, timescale=3),
                [14.1, 25.9, -34.9, -26.1], 250, 200,
                [-3, 3], cols="RdYlBu", log=False, clabel="SPI",
                save="GPCCv2020_025_SPI12_201912")  # -1 => last
# datapoint (2019-12)

N = 12
gm.discr_gridplot(minyrds.sel(timescale=N), 'decade',
                  [14.1, 24.9, -35.4, -26.1], width=125, height=120,
                  legtitle="decade", cols=my15,
                  save="GPCCv2020_025_SPI" + str(N) + "_mindecade")

gm.cts_facetplot(SWSA.SPI.isel(time=-1), 'timescale', ncols=3, vlims=[-3, 3],
                 bounds=[14.1, 25.9, -34.9, -26.1], width=500, height=500,
                 clabel="SPI", cols="RdYlBu")

Augs = np.array(["2017-08-01", "2018-08-01", "2019-08-01"],
                dtype="datetime64")  # end of August for 3 years

gm.cts_facetplot(SWSA.SPI.sel(time=Augs, timescale=36), 'time', ncols=3,
                 vlims=[-3, 3], bounds=[14.9, 23, -34.9, -26.9],
                 width=350, height=150, save="GPCCv2020_SPI36_2017-2019.png",
                 clabel="SPI36", cols="RdYlBu")

gm.cts_facetplot(SWSA.SPI.isel(time=-1).sel(timescale=[12, 24, 36, 60]),
                 'timescale', ncols=4, vlims=[-3, 3],  clabel="SPI",
                 bounds=[14.1, 25.9, -34.9, -26.1], width=350, height=120,
                 save="GPCCv2020_Dec2019_SPI12-60.png", cols="RdYlBu")

# %%
Augs = np.array(["2015-08-01", "2016-08-01", "2017-08-01", "2018-08-01"],
                dtype="datetime64")  # end of August for 1 year drought

# ensure date labels are accurate
SWSA_Aug12 = SWSA.sel(time=Augs, timescale=12)
SWSA_Aug12['time'] = SWSA_Aug12['time'] + np.timedelta64(30, 'D')

gm.cts_facetplot(SWSA_Aug12.SPI, 'time', ncols=4,
                 vlims=[-3, 3], bounds=[14.9, 23, -34.9, -26.9],
                 width=360, height=130, save="GPCCv2020_SPI12_2015-2018.png",
                 clabel="SPI12", cols="RdYlBu")
