#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr  7 14:25:41 2020

Determine various climatological seasonality properties of GPCC climatology
dataset over South-Western Southern Africa (SWSA)

@author: stefaan
"""
# %% imports

import xarray as xr
import numpy as np
import fn.xrclim_fns as xrc
import seaborn as sns
import gridfn.read as grd
import gridfn.maps as gm

sns.set_context("notebook")
sns.set_style("darkgrid")

# %% constants

# default colours to use for discrete variable mapping
discrete_colours = [
    "c",
    "g",
    "lime",
    "gold",
    "darkorange",
    "purple",
    "crimson",
    "dimgrey",
    "black",
]
hot5 = ["c", "black", "red", "darkorange", "yellow"]
paired9 = sns.color_palette("Paired_r", n_colors=10)[1:9]

# add only one blue

paired9.insert(0, sns.color_palette("Paired", n_colors=10)[1])

# my adaptation of Tol_muted for when first colour is to be reserved for the
# ocean, 2nd colour for single entity and other colours come in 2s; last
# colour for NA

# cyan black indigo wine sand olive green teal pale_grey
Tol_seafirst_paired = [
    "#88CCEE",
    "#000000",
    "#332288",
    "#882255",
    "#DDCC77",
    "#999933",
    "#117733",
    "#44AA99",
    "#DDDDDD",
]


# %% set-up

datadir = "../data/"
file = datadir + "gpcc_normals_025.nc"

# %% SH Midlat clim

rain_clim = xr.open_dataset(file, decode_times=False)  # problems with time
rain_clim["month"] = rain_clim.time + 1
rain_clim["annual"] = rain_clim.precip.sum(dim="time")

rain_clim["AMJJAS"] = rain_clim.precip.isel(time=slice(3, 9)).sum(dim="time")
rain_clim["JJ"] = rain_clim.precip.isel(time=slice(5, 7)).sum(dim="time")
rain_clim["JJA"] = rain_clim.precip.isel(time=slice(5, 8)).sum(dim="time")
rain_clim["SON"] = rain_clim.precip.isel(time=slice(8, 11)).sum(dim="time")
rain_clim["MAM"] = rain_clim.precip.isel(time=slice(2, 5)).sum(dim="time")
rain_clim["AM"] = rain_clim.precip.isel(time=slice(2, 4)).sum(dim="time")
rain_clim["AS"] = rain_clim.precip.isel(time=slice(7, 9)).sum(dim="time")
rain_clim["AMJ"] = rain_clim.precip.isel(time=slice(3, 6)).sum(dim="time")
rain_clim["JAS"] = rain_clim.precip.isel(time=slice(6, 9)).sum(dim="time")
rain_clim["DJF"] = rain_clim.precip.isel(time=slice(0, 2)).sum(
    dim="time"
) + rain_clim.precip.isel(time=slice(11, 12)).sum(dim="time")
rain_clim["ONDJFM"] = rain_clim.annual - rain_clim.AMJJAS
rain_clim["WvS"] = rain_clim.AMJJAS / rain_clim.ONDJFM
rain_clim["JJAvDJF"] = rain_clim.JJA / rain_clim.DJF
rain_clim["shoulder"] = rain_clim.AM + rain_clim.AS
rain_clim["coreVshoulder"] = 2 * (rain_clim.JJ) / (rain_clim.shoulder)
rain_clim["autumnVspring"] = (rain_clim.MAM) / (rain_clim.SON)
rain_clim["earlyVlate"] = rain_clim.AMJ / rain_clim.JAS
rain_clim["WvYr"] = rain_clim.AMJJAS / rain_clim.annual

# %% time-varying SWSA data

SWSA = (grd.read_time_man(datadir +
                          "gpcc_V2020_SWSA_025deg_14-26E_-35.5--26N.nc")).sel(
    lat=slice(-22, -35)
)
SWSA["ew_v_ann"] = xrc.ia_ew_tot(SWSA.precip, yrvar="time.year") / xrc.ia_tot(
    SWSA.precip, yrvar="time.year"
)

SWSA["ew_cont_q5"] = SWSA.ew_v_ann.quantile(q=0.05, dim="year")

SWSA["ew_cont_q95"] = SWSA.ew_v_ann.quantile(q=0.95, dim="year")

ann = SWSA.precip.groupby("time.year").sum(dim="time")
SWSA["clim"] = ann.mean(dim="year")
SWSA["sd"] = ann.std(dim="year")
SWSA["cv"] = SWSA.sd / SWSA.clim

SWSA_clim = rain_clim.sel(lat=slice(-22, -35), lon=slice(14, 27))

# %% GPCC v2018 domains

# domain names: simple
dnames = ["1-core", "2-periphery", "3-transition", "4-surroundings"]

SWSA["domain"] = dnames[0]

# core; default to periphery
SWSA["domain"] = SWSA.domain.where(
    (SWSA.ew_cont_q5 > 0.58) & (SWSA_clim.WvYr > 0.73), other=dnames[1]
)

# periphery
SWSA["domain"] = SWSA.domain.where(
    (SWSA.domain.isin(dnames[0:1]))
    | ((SWSA.ew_cont_q5 > 0.52) & (SWSA_clim.WvYr > 0.64))
    | (SWSA_clim.WvYr > 0.72),
    other=dnames[3],
)  # default to surroundings

# surroundings; if not then transition
SWSA["domain"] = SWSA.domain.where(
    (SWSA.domain.isin(dnames[0:2])) | (SWSA_clim.WvYr <= 0.59), other=dnames[2]
)

# ocean
SWSA["domain"] = SWSA.domain.where(SWSA_clim.AM > 0, other="0-no land")

# domain names: extended
dnames = [
    "1-core",
    "2b-periphery_S",
    "2a-periphery_N",
    "3b-transition_S",
    "3a-transition_N",
    "4b-surroundings_S",
    "4a-surroundings_N",
]

SWSA["domain_ex"] = dnames[0]  # strictest

# core
SWSA["domain_ex"] = SWSA.domain_ex.where(
    (SWSA.ew_cont_q5 > 0.58) & (SWSA_clim.WvYr > 0.73), other=dnames[1]
)

# periphery southern
SWSA["domain_ex"] = SWSA.domain_ex.where(
    (SWSA.domain_ex.isin(dnames[0:1]))
    | (((SWSA.ew_cont_q5 > 0.5) & (SWSA_clim.WvYr > 0.64)) |
       (SWSA_clim.WvYr > 0.72))
    & (SWSA.clim > 350),
    other=dnames[2],
)

# periphery northern
SWSA["domain_ex"] = SWSA.domain_ex.where(
    (SWSA.domain_ex.isin(dnames[0:2]))
    | (((SWSA.ew_cont_q5 > 0.5) & (SWSA_clim.WvYr > 0.64)) |
       (SWSA_clim.WvYr > 0.72))
    & (SWSA.clim <= 350),
    other=dnames[5],
)

# Southern surroundings (YRZ)
SWSA["domain_ex"] = SWSA.domain_ex.where(
    (SWSA.domain_ex.isin(dnames[0:3])) |
    (SWSA_clim.WvYr <= 0.59), other=dnames[3]
)

# Northern surroundings (LSRZ)
SWSA["domain_ex"] = SWSA.domain_ex.where(
    (SWSA.domain_ex.isin(dnames[0:5])) | (SWSA_clim.autumnVspring < 1.3),
    other=dnames[6],
)

# Northern transition
others = dnames[0:3] + dnames[5:7]  # regions already resolved
SWSA["domain_ex"] = SWSA.domain_ex.where(
    (SWSA.domain_ex.isin(others)) | (SWSA_clim.autumnVspring < 1.3),
    other=dnames[4]
)

# ocean points
SWSA["domain_ex"] = SWSA.domain_ex.where(SWSA_clim.AM > 0, other="0-no land")
# null scatter doesn't work, so make oceans 'NA'

# %% domain plots

sns.set_context("notebook", font_scale=0.9)

gm.discr_gridplot(
    SWSA,
    "domain",
    [14.5, 25, -35, -26.5],
    width=120,
    height=115,
    save="GPCC_v2018_domain5simple",
    cols=hot5,
)

gm.discr_gridplot(
    SWSA,
    "domain_ex",
    [14.5, 25, -35, -26.5],
    width=120,
    height=115,
    save="GPCC_v2018_domain5_ex",
    cols=Tol_seafirst_paired,
)
