#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 28 14:35:53 2020

Plots associated with clustering (see fn/station_clusters.py) using basemap
for mapping clusters.

@author: stefaan
"""

import pandas as pd
import xarray as xr
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.basemap import Basemap
import seaborn as sns
from scipy.cluster.hierarchy import dendrogram, linkage
import fn.wind_plots as wp

sns.set(style="darkgrid")

# Qualitative colour palettes

# Hex lists of colours

# Tol colours:

# blue red green yellow cyan purple grey black
Tol_bright = [
    "#4477AA",
    "#EE6677",
    "#228833",
    "#CCBB44",
    "#66CCEE",
    "#AA3377",
    "#BBBBBB",
    "#000000",
]

# my alternatives, with a different grey, at the end  (no. 4) and with
# black first

# blue red green yellow cyan purple black grey
Tol_b2 = [
    "#4477AA",
    "#EE6677",
    "#228833",
    "#CCBB44",
    "#66CCEE",
    "#AA3377",
    "#000000",
    "#555555",
]

# blue red green grey yellow cyan purple black
Tol_b3 = [
    "#4477AA",
    "#EE6677",
    "#228833",
    "#555555",
    "#CCBB44",
    "#66CCEE",
    "#AA3377",
    "#000000",
]

Tol_b1 = [
    "#000000",
    "#EE6677",
    "#CCBB44",
    "#66CCEE",
    "#228833",
    "#555555",
    "#4477AA",
    "#AA3377",
]  # black red yellow cyan green grey blue purple

# high contrast

Tol_hc = ["#004488", "#DDAA33", "#BB5566", "#000000"]  # blue yellow red black

# orange blue cyan magenta red teal grey black
Tol_vibrant = [
    "#EE7733",
    "#0077BB",
    "#33BBEE",
    "#EE3377",
    "#CC3311",
    "#009988",
    "#BBBBBB",
    "#000000",
]

# rose indigo sand green cyan wine teal olive purple pale_grey black
Tol_muted = [
    "#CC6677",
    "#332288",
    "#DDCC77",
    "#117733",
    "#88CCEE",
    "#882255",
    "#44AA99",
    "#999933",
    "#AA4499",
    "#DDDDDD",
    "#000000",
]
# orange red rose purple wine indigo blue cyan teal green olive sand black
# pale_grey
my15 = [
    "#EE7733",
    "#CC3311",
    "#CC6677",
    "#AA4499",
    "#882255",
    "#332288",
    "#0077BB",
    "#88CCEE",
    "#44AA99",
    "#117733",
    "#999933",
    "#DDCC77",
    "#000000",
    "#DDDDDD",
]

# my adaptation of Tol_muted for when first colour is to be reserved for the
# ocean, 2nd colour for single entity and other colours come in 2s; last
# colour for NA

# cyan black indigo wine sand olive green teal pale_grey
Tol_seafirst_paired = [
    "#88CCEE",
    "#000000",
    "#332288",
    "#882255",
    "#DDCC77",
    "#999933",
    "#117733",
    "#44AA99",
    "#DDDDDD",
]

# light_blue orange light_yellow pink light_cyan mint pear olive pale_grey
# black
Tol_light = [
    "#77AADD",
    "#EE8866",
    "#EEDD88",
    "#FFAABB",
    "#99DDFF",
    "#44BB99",
    "#BBCC33",
    "#AAAA00",
    "#DDDDDD",
    "#000000",
]

# Wong 2011 (nature methods) colours:

# Black, Orange, Sky_Blue, Bluish_Green, Yellow, Blue, Vermillion,
# Reddish purple
Wong = [
    "#000000",
    "#E69D00",
    "#56B4E9",
    "#009E74",
    "#F0E442",
    "#0072B2",
    "#D55C00",
    "#CC79A7",
]

# IBM

IBM = ["#648FFF", "#785EF0", "#DC267F", "FE6100", "FFB000"]


# %%

figdir = "../figures/"


def station_dendo(dist, thres, labs, desc):
    """
    Dendogram of a station clustering

    Parameters
    ----------
    dist : array or xarray DataArray
        Variables by which the distance should be computed. Should be
        transposed relative to the input of `station_clusters.clust_stations()`
    thres : float
        Distance beyond which to split clusters.
    labs : list of str
        Labels to place on the dendogram.
    desc : str
        Description for saving to file.

    Returns
    -------
    None.

    """
    linked = linkage(dist, "ward")
    plt.figure(figsize=(20, 13))
    dendrogram(
        linked,
        color_threshold=thres,
        orientation="top",
        labels=labs,
        leaf_rotation=90,
        distance_sort="descending",
        show_leaf_counts=True,
    )
    plt.savefig(figdir + desc + ".png", format="png", dpi=300)


# %%


def cluster_map(
    dist,
    n_clusters,
    bmap,
    colours,
    labs,
    lons,
    lats,
    desc,
    legloc="lower left",
    size=50,
):
    """
    Map of clusters on a basemap.

    This uses an instance of `basemap_plots.discrete_var_map()`. See this
    functions docstring for more information.

    Parameters
    ----------
    dist : xarray DataArray or array
        The variables to cluster on
    n_clusters : int
        Number of clusters to find
    bmap : Basemap
        The basemap to plot onto, returned by `station_basemap()`.
    colours : list, or str optional
        list or list-like of colour names or string defining a built-in
        seaborn colourmap. The default is Tol_bright [defined in this script].
    labs : list or array of str
        The levels of the mapped discrete variable as they are to be labelled
        in the legend.
    lons : list of float
        Longitude of each station (or equivalent point location) in `dist`.
        Must be the same length as the "station" (or equivalent) dimension
        in `dist`.
    lats : list of float
        Latitude of each station (or equivalent point location) in `dist`.
        Must be the same length as the "station" (or equivalent) dimension
        in `dist`.
    desc : str
        Description of data to use for saving the plot.
    legloc : str in {'best', 'upper right', 'upper left', 'lower left',
                       'lower right', 'right', 'center left', 'center right',
                       'lower center', 'upper center', 'center'}; or
            pair of floats; or 'int' in [0, 10], optional
        Where the legend is to be located. The default is 'lower left'. See
        `matplotlib.pyplot.scatter` parameter `loc` for more details.
        The default is "lower left".
    size : float, optional
        Size of circles to denote stations on the map. The default is 50.

    Returns
    -------
    None.

    """

    m = bmap

    colours = sns.color_palette(colours, n_colors=n_clusters)

    clust = dist.labels_

    p = dict()
    # store all the scatter plots so they can be removed separately from
    # the basemap
    for i in range(n_clusters):
        loc = clust == i
        p[i] = m.scatter(
            lons[loc],
            lats[loc],
            latlon=True,
            marker="o",
            s=size,
            zorder=1,
            color=colours[i],
            label=labs[i],
            edgecolors="black",
            linewidth=1,
            alpha=0.8,
        )

    plt.legend(loc=legloc)

    plt.savefig(
        figdir + desc + str(n_clusters) + ".png",
        format="png",
        dpi=150,
        bbox_inches="tight",
    )

    for i in range(n_clusters):
        # remove the scatter plots/clear the basemap foreground for next plot
        p[i].remove()


# %%


def seasonality_plot(
    clusternums,
    climdata,
    regnames,
    desc,
    normclim=True,
    var="rainfall",
    units=" (mm/mon)",
    stationdim="feature",
    colours=Tol_bright,
):
    """
    Line plot of mean monthly rainfall (or similar) per cluster.

    Data should be provided for a list of stations or other point locations.
    All list or array parameters should provide data for the same set of
    stations in the same sequence.

    Parameters
    ----------
    clusternums : list or array-like of int
       The cluster to which each station belongs, identified by a unique
       integer for each cluster. Usually the attribute 'labels_' of the output
       of `station_clusters.clust_stations()`.
    climdata : 2-D xarray DataArray (or similar)
        Monthly climatological data for the variable in question at each
        station. Dimension 1: along months, dimension 2: along stations.
    regnames : list of str
        Names to use for the clusters in the legend.
    desc : str
        Description of data to use for saving the plot.
    normclim : bool, optional
        Should the variable be normalised by annual totals for plotting?
        The default is True.
    var : str, optional
        Name to use for the plotted variable on the y-axis.
        The default is "rainfall".
    units : str, optional
        Units to use on the y-axis. The default is " (mm/mon)".
    stationdim : str, optional
        Name of the dimension in `climdata` along which stations are arranged.
        The default is "feature".
    colours : list, or str optional
        list or list-like of colour names or string defining a built-in
        seaborn colourmap. The default is Tol_bright [defined in this script].

    Returns
    -------
    None.

    """
    cluster_n = max(clusternums)              # last cluster = n_clusters-1

    station_num = climdata.station.argsort()  # for station cluster selection

    # initialise with the first (0th) cluster
    Df = (
        climdata.where(station_num.isin(np.where(clusternums == 0)))
        .mean(dim=stationdim)
        .to_dataframe()
    )
    Df["regime"] = regnames[0]

    # add the other clusters 1 by 1
    for i in range(cluster_n):
        Df2 = (
            climdata.where(station_num.isin(np.where(clusternums == (i + 1))))
            .mean(dim=stationdim)
            .to_dataframe()
        )
        Df2["regime"] = regnames[i + 1]  # i+1 because 0 has already been added
        Df = Df.append(Df2)
    Df = Df.reset_index()

    sns.set(style="darkgrid")
    sns.set_context("notebook")
    sns.set_palette(sns.color_palette(colours))

    varname = climdata.name     # name of the variable

    p = sns.relplot(
        x="month",
        y=varname,
        hue="regime",
        kind="line",
        data=Df,
        aspect=2,
        height=3.5,
        linewidth=2,
    )

    plt.xlabel("month")
    if normclim:
        plt.ylabel("proportion of annual " + var)
    else:
        plt.ylabel("mean monthly " + var + units)

    p.savefig(
        figdir + desc + "_" + str(cluster_n + 1) + "normclim" +
        str(normclim) + ".png",
        format="png",
        dpi=100,
    )


# %%


def variability_plot(
    datavar,
    labels,
    labsvar,
    desc,
    varname="rain",
    stationdim="station",
    var="rainfall",
    units=" (mm/yr)",
    std=True,
    colours=Tol_bright,
):
    """
    Time series plots of mean annual anomalies by cluster.

    Parameters
    ----------
    datavar : xarray DataArray
        Data to plot. Must either be standardised annual anomalies
        [if `std=True`] or raw rainfall data [if `std=False`].
        Must have either a "year" [if `std=True`] or "time" [if `std=False`]
        and a `stationdim` dimension.
    labels : list or array-like of int
        Same length as the `stationdim` in `datavar`. Which cluster each
        station belongs to. Usually the attribute 'labels_' of the output
        of `station_clusters.clust_stations()`.
    labsvar : list or array of str
        Names of the clusters to use in the legend, in the order of `labels`.
    desc : str
        Description of data to use for saving the plot.
    stationdim : str, optional
        Name of the dimension in `datavar` along which stations are arranged.
        The default is "station".
    var : str, optional
        Name to use for the plotted variable on the y-axis.
        The default is "rainfall".
    units : str, optional
        Units to use on the y-axis. The default is " (mm/yr)". Ignored if
        `std=True`
    std : bool, optional
        Are the supplied data in `datavar` standardised annual anomalies, or
        should annual totals be computed? The default is True.
    colours : list, or str optional
        list or list-like of colour names or string defining a built-in
        seaborn colourmap. The default is Tol_bright [defined in this script].

    Returns
    -------
    None.

    """
    varname = datavar.name
    S = datavar[stationdim]
    n_clusters = max(labels) + 1  # number of clusters
    Df0 = datavar.where(S.isin(S[np.where(labels == (n_clusters - 1))]),
                        drop=True)
    if std:
        Df = Df0.groupby("year").mean(dim=(stationdim)).to_dataframe()
        Df["region"] = labsvar[n_clusters - 1]
        for i in range(n_clusters - 1):
            Df0 = (
                datavar.where(S.isin(S[np.where(labels == i)]))
                .groupby("year")
                .mean(dim=(stationdim))
                .to_dataframe()
            )
            Df0["region"] = labsvar[i]
            Df = Df.append(Df0)

    else:  # only works if data have a "time" dimension
        Df0.name = varname
        Df = Df0.groupby("time.year").mean(dim=("time",
                                                stationdim)).to_dataframe()
        Df["region"] = labsvar[n_clusters - 1]
        for i in range(n_clusters - 1):
            Df0 = (
                datavar.where(S.isin(S[np.where(labels == i)]))
                .groupby("time.year")
                .mean(dim=("time", stationdim))
                .to_dataframe()
            )
            Df0["region"] = labsvar[i]
            Df = Df.append(Df0)

    sns.set(style="darkgrid")
    sns.set_context("notebook")
    sns.set_palette(sns.color_palette(colours))

    Df["year"] = Df.index

    fig = sns.relplot(x="year", y=varname, hue="region", kind="line", data=Df,
                      aspect=2)
    plt.xlabel("year")
    plt.ylabel("Standardised " * std + var + units * (1 - std))
    fig.savefig(figdir + "std_" * std + str(n_clusters) + desc + ".png")


# %%
def contrib_plot(
    datavar,
    labels,
    labsvar,
    descr,
    ncol=2,
    stationdim="station",
    contribdim="CWT",
    addmu=True,
    width=200,
    height=200,
    colours=Tol_bright,
):
    """
    Bar plots of the contribution of CWTs to annual rainfall by cluster.

    Each cluster is plotted as a facet with bars for each CWT or similar
    categorical variable.

    Parameters
    ----------
    datavar : xarray DataArray
        Contribution data: the climatological proportion of rainfall
        (or similar variable) occurring under each unique value of
        `contribdim`. Must be 2-D data with `contribdim` and `stationdim`.
    labels : list or array-like of int
        Same length as the `stationdim` in `datavar`. Which cluster each
        station belongs to. Usually the attribute 'labels_' of the output
        of `station_clusters.clust_stations()`.
    labsvar : list or array of str
        Names of the clusters to use in the legend, in the order of `labels`.
    desc : str
       Description of data to use for saving the plot.
    ncol : int, optional
        Number of columns in the output facetplot. The default is 2.
    stationdim : str, optional
        Name of the dimension in `datavar` along which stations are arranged.
        The default is "station".
    contribdim : str, optional
        Name of the dimension along which the categories for which rainfall is
        aggregated are arranged. The default is "CWT".
    addmu : bool, optional
        Should a panel be added to the facetgrid for the overall spatial mean
        contributions? The default is True.
    width : float, optional
        Width in mm of the full figure. The default is 200.
    height : float, optional
        Height in mm of the full figure. The default is 200.
    colours : list, or str optional
        list or list-like of colour names or string defining a built-in
        seaborn colourmap. The default is Tol_bright [defined in this script].

    Returns
    -------
    None.

    """
    S = datavar[stationdim]

    n_clusters = max(labels) + 1

    # data for the first (reference) cluster
    da = datavar.where(S.isin(S[np.where(labels == 0)]),
                       drop=True).mean(dim=stationdim)
    da["region"] = labsvar[0]
    for i in range(n_clusters - 1):
        da0 = datavar.where(S.isin(S[np.where(labels == (i + 1))])
                            ).mean(dim=stationdim)
        da0["region"] = labsvar[i + 1]
        da = xr.concat([da, da0], dim="region")

    if addmu:
        da0 = datavar.mean(dim=stationdim)
        da0["region"] = "Mean"
        da = xr.concat([da, da0], dim="region")

    # plotting
    wp.contrib_bars(
        da,
        "region",
        "wdirn_clusts" + str(n_clusters) + descr,
        ncol,
        width=width,
        height=height,
        colours=colours,
    )
