#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 28 14:23:18 2020

@author: stefaan
"""
import pandas as pd
import xarray as xr
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.basemap import Basemap
import matplotlib.colors as colors
import fn.xrclim_fns as xrc
import fn.station_clusters as sc
import seaborn as sns

sns.set(style="darkgrid")

FigDir = "../figures/"

# Qualitative colour palettes

# Hex lists of colours

# Tol colours:

# blue red green yellow cyan purple grey black
Tol_bright = [
    "#4477AA",
    "#EE6677",
    "#228833",
    "#CCBB44",
    "#66CCEE",
    "#AA3377",
    "#BBBBBB",
    "#000000",
]

# my alternatives, with a different grey, at the end  (no. 4) and with
# black first

# blue red green yellow cyan purple black grey
Tol_b2 = [
    "#4477AA",
    "#EE6677",
    "#228833",
    "#CCBB44",
    "#66CCEE",
    "#AA3377",
    "#000000",
    "#555555",
]

# blue red green grey yellow cyan purple black
Tol_b3 = [
    "#4477AA",
    "#EE6677",
    "#228833",
    "#555555",
    "#CCBB44",
    "#66CCEE",
    "#AA3377",
    "#000000",
]

Tol_b1 = [
    "#000000",
    "#EE6677",
    "#CCBB44",
    "#66CCEE",
    "#228833",
    "#555555",
    "#4477AA",
    "#AA3377",
]  # black red yellow cyan green grey blue purple

# high contrast

Tol_hc = ["#004488", "#DDAA33", "#BB5566", "#000000"]  # blue yellow red black

# orange blue cyan magenta red teal grey black
Tol_vibrant = [
    "#EE7733",
    "#0077BB",
    "#33BBEE",
    "#EE3377",
    "#CC3311",
    "#009988",
    "#BBBBBB",
    "#000000",
]

# rose indigo sand green cyan wine teal olive purple pale_grey black
Tol_muted = [
    "#CC6677",
    "#332288",
    "#DDCC77",
    "#117733",
    "#88CCEE",
    "#882255",
    "#44AA99",
    "#999933",
    "#AA4499",
    "#DDDDDD",
    "#000000",
]
# orange red rose purple wine indigo blue cyan teal green olive sand black
# pale_grey
my15 = [
    "#EE7733",
    "#CC3311",
    "#CC6677",
    "#AA4499",
    "#882255",
    "#332288",
    "#0077BB",
    "#88CCEE",
    "#44AA99",
    "#117733",
    "#999933",
    "#DDCC77",
    "#000000",
    "#DDDDDD",
]

# my adaptation of Tol_muted for when first colour is to be reserved for the
# ocean, 2nd colour for single entity and other colours come in 2s; last
# colour for NA

# cyan black indigo wine sand olive green teal pale_grey
Tol_seafirst_paired = [
    "#88CCEE",
    "#000000",
    "#332288",
    "#882255",
    "#DDCC77",
    "#999933",
    "#117733",
    "#44AA99",
    "#DDDDDD",
]

# light_blue orange light_yellow pink light_cyan mint pear olive pale_grey
# black
Tol_light = [
    "#77AADD",
    "#EE8866",
    "#EEDD88",
    "#FFAABB",
    "#99DDFF",
    "#44BB99",
    "#BBCC33",
    "#AAAA00",
    "#DDDDDD",
    "#000000",
]

# Wong 2011 (nature methods) colours:

# Black, Orange, Sky_Blue, Bluish_Green, Yellow, Blue, Vermillion,
# Reddish purple
Wong = [
    "#000000",
    "#E69D00",
    "#56B4E9",
    "#009E74",
    "#F0E442",
    "#0072B2",
    "#D55C00",
    "#CC79A7",
]

# IBM

IBM = ["#648FFF", "#785EF0", "#DC267F", "FE6100", "FFB000"]

# %% Functions


def station_basemap(bounds, width=200, height=200):
    """
    Draws the basemap for other functions to plot over later to save on RAM.

    Figure width and height are to be provided in mm.

    Parameters
    ----------
    bounds : length-4 list or array-like of floats
        A list = [lonW, lonE, latS, latN], where lonW and lonE are
        the longitudes on the western and eastern edges, respectively of the
        rectangular map, and latS and latN the latitudes at its respective
        southern and northern edges.
    width : float, optional
        Width of figure in mm. The default is 200.
    height : float, optional
        Height of figure in mm. The default is 200.

    Returns
    -------
    m1 : Basemap
        The map to plot on
    fig : matplotlib Figure
        Figure object containg the Basemap.

    """
    lonW = bounds[0]
    lonE = bounds[1]
    latS = bounds[2]
    latN = bounds[3]

    fig = plt.figure(figsize=(width / 25, height / 25))  # ~25 mm => in
    pl = fig.add_subplot(1, 1, 1)  # add the ith subplot; i+1

    m1 = Basemap(
        projection="cea",
        resolution="i",
        lat_0=(latS + latN) / 2,
        lon_0=(lonW + lonE) / 2,
        llcrnrlat=latS,
        urcrnrlat=latN,
        llcrnrlon=lonW,
        urcrnrlon=lonE,
        epsg=4269,
    )

    griddelta = max(0.5, np.floor((latN - latS) / 6),
                    np.floor((lonE - lonW) / 6))

    m1.drawparallels(
        np.arange(np.ceil(latS), np.ceil(latN), griddelta),
        labels=[1, 0, 0, 0],
        dashes=[1, 1],
        linewidth=0.25,
        color="0.5",
    )  # label lat lines left;
    # labels: [left, right, top, bottom]
    m1.drawmeridians(
        np.arange(np.ceil(lonW), np.ceil(lonE), griddelta),
        labels=[0, 0, 1, 0],
        dashes=[1, 1],
        linewidth=0.25,
        color="0.5",
    )
    m1.shadedrelief()
    m1.drawcountries(color="black", linewidth=3)

    fig.subplots_adjust(
        left=0.1, right=0.95, top=0.93, bottom=0.01, hspace=0.0, wspace=0.08
    )
    # cbaxes = fig.add_axes([0.9, 0.15, 0.035, 0.7])

    # plt.colorbar(sc, cax = cbaxes)
    return m1, fig


def station_basemap_fg(
    bounds,
    faceter,
    nrow=2,
    ncol=2,
    width=200,
    height=200,
    titles=None,
    nlats=6,
    nlons=6,
    lyt="std",
):
    """
    Facetgrid version of `station_basemap()`

    Parameters
    ----------
    bounds : length-4 list or array-like of floats
       A list = [lonW, lonE, latS, latN], where lonW and lonE are
       the longitudes on the western and eastern edges, respectively of the
       rectangular map, and latS and latN the latitudes at its respective
       southern and northern edges.
    faceter : str
        Variable to facet by.
    nrow : int,  optional
        Number of rows in the facetgrid. The default is 2.
    ncol : int, optional
        Number of columns in the facetgrid. The default is 2.
    width : float, optional
        Width of the full facetgrid in mm. The default is 200.
    height : float, optional
        Height of the full facetgrid in mm. The default is 200.
    titles : list or list-like of str, optional
        List as long as the number of facets in the grid. If set to None
        (default)
    nlats : int, optional
        Number of latitude points to label (approximate). The default is 6.
    nlons : int, optional
       Number of longitude points to label (approximate). The default is 6.
    lyt : str, optional
        Plot layout. Can be set to 'constrained' for a constrained layout or
        'tight' for a tight layout, but note that both are experimental.
        Any other specification leads to a std plot. Default is 'std'.
        The default is 'std'.

    Returns
    -------
    m1 : Basemap
        Map to plot points on.
    fig : Matplotlib figure
        The faceted figure object to plot onto.
    pl : dictionary of subplot axes
        Named dictionary for the facet subplots.

    """
    lonW = bounds[0]
    lonE = bounds[1]
    latS = bounds[2]
    latN = bounds[3]

    N = nrow * ncol  # total subplots

    if titles is None:
        titles = np.repeat("", N)
        lon_labs = [0, 0, 1, 0]  # position longitude labels on top
    else:
        lon_labs = [0, 0, 0, 1]

    fig = plt.figure(
        figsize=(width / 25, height / 25),  # 25 mm => in
        constrained_layout=lyt == "constrained",
    )

    fig.set_tight_layout(lyt == "tight")

    pl = {}  # dictionary to fill with subplot axes

    lat0 = (latS + latN) / 2
    lon0 = (lonW + lonE) / 2
    ydelta = max(0.5, np.floor((latN - latS) / nlats))
    xdelta = max(0.5, np.floor((lonE - lonW) / nlons))

    for i, seas in enumerate(faceter):

        if lon_labs[2] == 0:  # on bottom row only
            lonlabs = lon_labs * (i >= (nrow - 1) * (N / nrow))
        else:                 # on top row only
            lonlabs = lon_labs * (i <= (N / nrow))

        pl[seas] = fig.add_subplot(nrow, ncol, i + 1)  # add the ith subplot
        pl[seas].title.set_text(titles[i])

        m1 = Basemap(
            projection="cea",
            resolution="i",
            lat_0=lat0,
            lon_0=lon0,
            llcrnrlat=latS,
            urcrnrlat=latN,
            llcrnrlon=lonW,
            urcrnrlon=lonE,
            epsg=4269,
        )

        m1.drawparallels(
            np.arange(np.ceil(latS), np.ceil(latN), ydelta),
            labels=[i % ncol == 0, 0, 0, 0],
            dashes=[1, 1],
            linewidth=0.25,
            color="0.5",
        )  # label lat lines left;
        # label only if in the left-hand most column
        # labels: [left, right, top, bottom]
        m1.drawmeridians(
            np.arange(np.ceil(lonW), np.ceil(lonE), xdelta),
            labels=lonlabs,
            dashes=[1, 1],
            linewidth=0.25,
            color="0.5",
        )
        m1.shadedrelief()
        m1.drawcountries(color="black", linewidth=3)

    # fig.subplots_adjust(left=0.05, right=0.85, top=0.83, bottom=0.01,
    #                         hspace=0.1, wspace=0.12)

    return m1, fig, pl


def discrete_var_map(
    data,
    descr,
    clabel,
    bmap,
    var=None,
    colours=Tol_bright,
    legendloc="lower left",
    add_cols=0,
    coli=0,
):
    """
    Map a discrete variable at point locations with coloured scatter circles.

    Maps onto a Basemap object returned by `station_basemap()` or similar.

    Parameters
    ----------
    data : xarray DataArray, Dataset or pandas DataFrame
        Object containing data to plot
    descr : str
        Description of data to use for saving the plot.
    clabel : str
        Label for the colourbar.
    bmap : Basemap
        The basemap to plot onto, returned by `station_basemap()`.
    var : None or str, optional
       If a xarray Dataset or pandas DataFrame is passed, the name of the
       variable to be plotted. Pass None if `data` is a DataArray.
       The default is None.
    colours : list, or str optional
        list or list-like of colour names or string defining a built-in
        seaborn colourmap. The default is Tol_bright [defined in this script].
    legendloc : str in {'best', 'upper right', 'upper left', 'lower left',
                       'lower right', 'right', 'center left', 'center right',
                       'lower center', 'upper center', 'center'}; or
            pair of floats; or 'int' in [0, 10], optional
        Where the legend is to be located. The default is 'lower left'. See
        `matplotlib.pyplot.scatter` parameter `loc` for more details.
        The default is "lower left".
    add_cols : int, optional
        Applies to pre-defined seaborn colour maps with variable length.
        By default, the number of colours selected from the scheme is equal
        to the number of unique values of the discrete plotting variable.
        This number can be increased using this parameter. `add_cols` > `coli`
        otherwise an error results. The default is 0.
    coli : int, optional
        Point in colourmap to start selecting from. This applies only to
        pre-defined seaborn colourmaps. The default is 0.

    Raises
    ------
    Exception
        If coli > add_cols too few colours will be selected.
        Or if coli > the length of the selected colourscheme.
    Warning
        If colours need to be recycled as fewer are supplied than there are
        unique values of the categorical variable plotted.

    Returns
    -------
    None.

    """
    _lons = data.lon
    _lats = data.lat

    if var is not None:
        data = data[var]

    m = bmap

    vals = data.values
    nvals = len(list(set(vals)))

    # check only applies if a built-in colour scheme name is provided
    if isinstance(colours, str):
        if coli > add_cols:
            raise Exception(
                "Please ensure that coli <= add_cols, otherwise"
                + "insufficiently many colours are available."
            )
        else:
            colours = sns.color_palette(colours[coli:],
                                        n_colors=nvals + add_cols)

    if not isinstance(colours, str):
        if len(colours) < coli:
            raise Exception("Not enough colours to start from " + str(coli))

        if nvals > len(colours) - coli:
            raise Warning(
                "Insufficiently many colours provided "
                + "once requested colours were removed; "
                + "note that colours will be recycled"
            )

        colours = sns.color_palette(colours[coli:])

    sc = dict()

    for i in range(nvals):
        value = pd.unique(vals)[i]
        loc = vals == value
        sc[i] = m.scatter(
            _lons[loc],
            _lats[loc],
            latlon=True,
            marker="o",
            s=90,
            zorder=1,
            color=colours[i],
            label=value,
            edgecolors="black",
            linewidth=2,
            alpha=0.9,
        )

    plt.legend(loc=legendloc, title=clabel)

    plt.savefig(FigDir + descr + ".png", format="png", dpi=300,
                bbox_inches="tight")
    for i in range(nvals):
        sc[i].remove()


def cts_var_map(
    data,
    descr,
    clabel,
    bmap,
    fig,
    var=None,
    colours="plasma_r",
    log=False,
    vlims=None,
    eps=0.01,
):
    """
    Maps a continuous variable to coloured scatter points on a Basemap.

    Parameters
    ----------
    data : xarray DataArray, Dataset or pandas DataFrame
        Object containing data to plot
    descr : str
        Description of data to use for saving the plot.
    clabel : str
        Label for the colourbar.
    bmap : Basemap
        The basemap to plot onto, returned by `station_basemap()`.
    fig : Matplotlib figure
        Figure for axes to plot on, as returned by `station_basemap()`.
    var : None or str, optional
       If a xarray Dataset or pandas DataFrame is passed, the name of the
       variable to be plotted. Pass None if `data` is a DataArray.
       The default is None.
    colours : str, optional
        Continuous colour map. The default is "plasma_r".
    log : bool, optional
        Should a log-scale be applied to the colour scale?
        The default is False.
    vlims : tuple or length-2 list or list-like, optional
        minimum and maximum values to use for the colour scale. If None is
        supplied, these are deduced from the data. The default is None.
    eps : float, optional
        A small value to add to data if a log-scale is used and the minimum
        values is exactly 0. The default is 0.01.

    Raises
    ------
    ValueError
        If `log=True`, but there are negative values in the plotting variable.

    Returns
    -------
    None.

    """
    _lons = data.lon
    _lats = data.lat

    if var is not None:
        data = data[var]

    if vlims is None:
        vlims = [data.min(), data.max()]

    m = bmap
    # m.drawmapboundary(fill_color='white', color="gray")

    extend = "neither"  # default to neither end of colourbar being extended
    if log:
        if data.min() < 0:
            raise ValueError("cannot use log scale on variable with negatives")
        elif data.min() == 0:
            data = data + eps
            cnorm = colors.LogNorm(vmin=0.1, vmax=vlims[1] + eps, clip=True)
            extend = "min"
        else:
            cnorm = colors.LogNorm(vmin=vlims[0], vmax=vlims[1], clip=True)
    else:
        cnorm = colors.Normalize(vmin=vlims[0], vmax=vlims[1], clip=True)

    if data.min() < vlims[0]:
        extend = "min"
    if data.max() > vlims[0]:
        if extend == "min":
            extend = "both"
        else:
            extend = "max"

    sc = m.scatter(
        _lons,
        _lats,
        latlon=True,
        marker="o",
        s=60,
        zorder=1,
        norm=cnorm,
        c=data,
        cmap=colours,
        edgecolors="black",
        linewidth=2,
        alpha=0.9,
    )

    # colour bar
    plt.legend("", framealpha=0)

    #    fig.subplots_adjust(left=0.05, right=0.85, top=0.93, bottom=0.01,
    #                        hspace=0.0, wspace=0.08)
    #    cbaxes = fig.add_axes([0.9, 0.15, 0.035, 0.7])

    cbar = plt.colorbar(mappable=sc, shrink=0.6, extend=extend)

    cbar.set_label(clabel)

    plt.savefig(FigDir + descr + ".png", format="png", dpi=300,
                bbox_inches="tight")
    cbar.remove()
    sc.remove()


def cts_var_map_fg(
    data,
    descr,
    clabel,
    bmap,
    fig,
    axdict,
    facetvar,
    var=None,
    colours="plasma_r",
    log=False,
    vlims=None,
):
    """
    facetgrid equivalent of `cts_var_map()`.

    This function takes inputs from the output of `station_basemap_fg`.

    Parameters
    ----------
    All parameters are as in `cts_var_map()` except

    bmap : Basemap
        As returned by `station_basemap_fg()`.
    fig : Matplotlib figure
        As returned by `station_basemap_fg()`.
    axdict : dict of subplot axes.
        As returned by `station_basemap_fg()`.
    facetvar : str
        Name of variable to facet by.

    Returns
    -------
    None.
    """
    _lons = data.lon
    _lats = data.lat

    if var is not None:
        data = data[var]

    if vlims is None:
        vlims = [data.min(), data.max()]

    m = bmap
    # m.drawmapboundary(fill_color='white', color="gray")

    extend = "neither"  # default to neither end of colourbar being extended
    if log:
        if data.min() < 0:
            raise ValueError("cannot use log scale on variable with negatives")
        elif data.min() == 0:
            data = data + 0.01
            cnorm = colors.LogNorm(vmin=0.1, vmax=vlims[1] + 0.01, clip=True)
            extend = "min"
        else:
            cnorm = colors.LogNorm(vmin=vlims[0], vmax=vlims[1], clip=True)
    else:
        cnorm = colors.Normalize(vmin=vlims[0], vmax=vlims[1], clip=True)

    if data.min() < vlims[0]:
        extend = "min"
    if data.max() > vlims[0]:
        if extend == "min":
            extend = "both"
        else:
            extend = "max"

    for seas in axdict.keys():
        seasvar = data.sel({facetvar: seas})
        sca = m.scatter(
            _lons,
            _lats,
            latlon=True,
            marker="o",
            s=60,
            zorder=1,
            norm=cnorm,
            c=seasvar,
            cmap=colours,
            edgecolors="black",
            linewidth=2,
            alpha=0.9,
            ax=axdict[seas],
        )

    # colour bar
    plt.legend("", framealpha=0)

    # colourbar width; wider relative to smaller plots
    cwidth = 0.01 + 0.3 / fig.get_size_inches()[0]  # width in inches

    fig.subplots_adjust(
        left=0.02,
        right=1 - (cwidth + 0.05),
        top=0.83,
        bottom=0.01,
        hspace=0.05,
        wspace=0.05,
    )

    cbax = fig.add_axes(
        [1 - (cwidth + 0.03), 0.09, cwidth, 0.65]
    )  # left, bottom, width, height

    cbar = plt.colorbar(sca, shrink=0.6, cax=cbax, extend=extend)
    cbar.set_label(clabel)

    plt.savefig(FigDir + descr + ".png", format="png", dpi=300,
                bbox_inches="tight")
    cbar.remove()
    sca.remove()


# %% seasonal contributions and comparisons


def plot_seas(
    da,
    descr,
    bmap,
    fig,
    months=np.arange(4, 10),
    seas="AMJJAS",
    vlims=[0, 100],
):
    """
    Special case of `cts_var_map()` for a seasonal proportion of rainfall.

    The default case is for "AMJJAS" (the Austral winter semester),
    i.e. the climatological proportion of annual rainfall
    (or similar variable) observed between April and September.

    Parameters
    ----------
    da : xarray DataArray
        Continuous variable, usually rainfall, to derive a monthly total
        climatology for.
    descr : str
        Description of data to use for saving the plot.
    bmap : Basemap
        The basemap to plot onto, returned by `station_basemap()`.
    fig : Matplotlib figure
        Figure for axes to plot on, as returned by `station_basemap()`.
    months : list or array of ints, optional
        The months to include in the season. The default is np.arange(4,10).
    seas : str, optional
        Name to use for the season. The default is 'AMJJAS'.
    vlims : tuple, or length-2 list or array, optional
        The minimum and maximum values to use for the colour scale.
        The default is [0,100].

    Returns
    -------
    None.

    """
    # normalised monthly climatology
    nclim = xrc.climatology(da)[1].sel(month=months).sum(dim="month")
    nclim.name = "proprain" + seas

    cts_var_map(
        data=nclim * 100,
        descr=nclim.name + descr,
        bmap=bmap,
        fig=fig,
        clabel="climatological % rainfall " + seas,
        log=False,
        vlims=vlims,
    )


def plot_seas_fg(
    da,
    descr,
    bmap,
    fig,
    axdict,
    seaslist=["DJF", "MAM", "JJA", "SON"],
    monthlist=[
        np.array([1, 2, 12]),
        np.arange(3, 6),
        np.arange(6, 9),
        np.arange(9, 12),
    ],
    vlims=[0, 100],
    colours="plasma",
):
    """
    Facetgrid version of `plot_seas()`.

    Plots a facet for each of the seasons supplied. Defaults to the four
    standard meteorological seasons [summer: DJF, autumn: MAM, winter: JJA,
    spring: SON, in the SH].

    Parameters
    ----------
    da : xarray DataArray
        Continuous variable, usually rainfall, to derive a monthly total
        climatology for.
    descr : str
        Description of data to use for saving the plot.
    bmap : Basemap
        As returned by `station_basemap_fg()`.
    fig : Matplotlib figure
        As returned by `station_basemap_fg()`.
    axdict : dict of subplot axes.
        As returned by `station_basemap_fg()`.
    seaslist : list or array of str, optional
        Names to use for the seasons plotted.
        The default is ['DJF', 'MAM', 'JJA', 'SON'].
    monthlist : list of lists or arrays, of ints, optional
        List containing the months in each of the seasons to use.
        Must be the same length as seaslist. The default is:
            [np.array([1,2,12]), np.arange(3,6), np.arange(6,9),
             np.arange(9,12)].
    vlims : tuple, or length-2 list or array, optional
        The minimum and maximum values to use for the colour scale.
        The default is [0,100].

    Returns
    -------
    None.

    """
    # normalised monthly climatology
    nclim = xrc.climatology(da)[1]

    ns = len(seaslist)
    months = monthlist[0]
    seas = seaslist[0]
    nseasclim = nclim.sel(month=months).sum(dim="month")
    nseasclim["season"] = seas
    for i in np.arange(1, ns):
        months = monthlist[i]
        seas = seaslist[i]
        tempclim = nclim.sel(month=months).sum(dim="month")
        tempclim["season"] = seas
        nseasclim = xr.concat([nseasclim, tempclim], dim="season")

    cts_var_map_fg(
        data=nseasclim * 100,
        descr="proprain_seas" + descr,
        bmap=bmap,
        fig=fig,
        axdict=axdict,
        facetvar="season",
        colours=colours,
        clabel="climatological % rainfall",
        log=False,
        vlims=vlims,
    )


# %% seasonal functions


def plot_seasratio(
    da,
    descr,
    bmap,
    fig,
    vlims=[0.25, 4],
    months_num=np.arange(4, 10),
    seas_num="AMJJAS",
    months_den=np.append(np.arange(1, 4), np.arange(10, 13)),
    seas_den="ONDJFM"
):
    """
    As with plot_seas, but for the ratio between two seasons.

    Defaults to AMJJAS:ONDJFM

    Parameters
    ----------
    As for
    vlims : tuple or length-2 array or list, optional
        Minimum and maximum values for colour scale. The default is [0.25, 4].
    months_num : list or array of int, optional
        Months making up the season used in the numerator.
        The default is np.arange(4,10).
    seas_num : str, optional
        Name for the season used in the numerator. The default is 'AMJJAS'.
    months_den : list or array of int, optional
        Months making up the season used in the denominator.
        The default is np.append(np.arange(1,4), np.arange(10,13)).
    seas_den : str, optional
        Name for the season used in the denominator. The default is 'ONDJFM'.

    Returns
    -------
    None.

    """
    nclim = xrc.climatology(da)[0]
    nclim_num = nclim.sel(month=months_num).sum(dim="month")  # numerator
    nclim_den = nclim.sel(month=months_den).sum(dim="month")  # denominator
    nclim = nclim_num / nclim_den
    nclim.name = seas_num + ":" + seas_den

    cts_var_map(
        data=nclim,
        descr=nclim.name + descr,
        bmap=bmap,
        fig=fig,
        clabel=nclim.name + " climatological ratio",
        log=True,
        vlims=vlims,
    )

# %% Discrete multiplot


def discr_multiplot(
    data,
    plotvars,
    descr,
    bounds,
    clabels,
    nrow=2,
    ncol=2,
    width=200,
    height=200,
    titles=None,
    nlats=6,
    nlons=6,
    lyt="std",
    colours=Tol_bright,
    legendloc="lower left",
    add_cols=0,
    coli=0,
):
    """
    Multiple discrete plots on the same axes.

    Roughly is to `discrete_var_map()` what  `cts_var_map_fg()` is to
    `cts_var_map()`, but plotting multiple variables rather than a proper
    "facetgrid".

    Parameters
    ----------
    data : xarray Dataset or pandas DataFrame
        Must contain longitude dimension/variable named 'lon', latitude
        dimension named 'lat' and all variables listed in plotvars.
    plotvars : list of str
        List of variables from `data` to be plotted.
    descr : See `discrete_var_map()`
    bounds : See `discrete_var_map()`
    clabels : list of str
        List of colourbar labels for the suplots.
    nrow : int, optional
        Rows of subplots. The default is 2.
    ncol : int, optional
        Columns of subplots. The default is 2.
    width : float, optional
        Width of full figure in mm. The default is 200.
    height : float, optional
        Height of full figure in mm. The default is 200.
    titles : list of str, optional
        Headings of the subplots. The default is None.
    nlats : int, optional
        Number of latitude points to label (approximate). The default is 6.
    nlons : int, optional
       Number of longitude points to label (approximate). The default is 6.
    lyt : str, optional
        Plot layout. Can be set to 'constrained' for a constrained layout or
        'tight' for a tight layout, but note that both are experimental.
        Any other specification leads to a std plot. Default is 'std'.
    colours : list, or str optional
        list or list-like of colour names or string defining a built-in
        seaborn colourmap. The default is Tol_bright [defined in this script].
    legendloc : str in {'best', 'upper right', 'upper left', 'lower left',
                       'lower right', 'right', 'center left', 'center right',
                       'lower center', 'upper center', 'center'}; or
            pair of floats; or 'int' in [0, 10], optional
        Where the legend is to be located. The default is 'lower left'. See
        `matplotlib.pyplot.scatter` parameter `loc` for more details.
        The default is "lower left".
    add_cols : int, optional
        Applies to pre-defined seaborn colour maps with variable length.
        By default, the number of colours selected from the scheme is equal
        to the number of unique values of the discrete plotting variable.
        This number can be increased using this parameter. `add_cols` > `coli`
        otherwise an error results. The default is 0.
    coli : int, optional
        Point in colourmap to start selecting from. This applies only to
        pre-defined seaborn colourmaps. The default is 0.

    Returns
    -------
    None.

    """
    _lons = data.lon
    _lats = data.lat

    lonW = bounds[0]
    lonE = bounds[1]
    latS = bounds[2]
    latN = bounds[3]

    N = nrow * ncol  # total subplots

    if len(plotvars) != N:
        raise Exception(
            "Please ensure that the number of variables to plot"
            + "equals the number of panels requested"
        )

    if titles is None:
        titles = np.repeat("", N)
        lon_labs = [0, 0, 1, 0]  # position longitude labels on top
    else:
        lon_labs = [0, 0, 0, 1]

    fig = plt.figure(
        figsize=(width / 25, height / 25),  # 25 mm => in
        constrained_layout=lyt == "constrained",
    )

    fig.set_tight_layout(lyt == "tight")

    pl = {}  # dictionary to fill with subplot axes

    lat0 = (latS + latN) / 2
    lon0 = (lonW + lonE) / 2
    ydelta = max(0.5, np.floor((latN - latS) / nlats))
    xdelta = max(0.5, np.floor((lonE - lonW) / nlons))

    for k, var in enumerate(plotvars):

        if lon_labs[2] == 0:    # on bottom row only
            lonlabs = lon_labs * (k >= (nrow - 1) * (N / nrow))
        else:                   # on top row only
            lonlabs = lon_labs * (k <= (N / nrow))

        pl[var] = fig.add_subplot(nrow, ncol, k + 1)  # add the kth subplot
        pl[var].title.set_text(titles[k])

        m1 = Basemap(
            projection="cea",
            resolution="i",
            lat_0=lat0,
            lon_0=lon0,
            llcrnrlat=latS,
            urcrnrlat=latN,
            llcrnrlon=lonW,
            urcrnrlon=lonE,
            epsg=4269,
        )

        m1.drawparallels(
            np.arange(np.ceil(latS), np.ceil(latN), ydelta),
            labels=[k % ncol == 0, 0, 0, 0],
            dashes=[1, 1],
            linewidth=0.25,
            color="0.5",
        )  # label lat lines left;
        # label only if in the left-hand most column
        # labels: [left, right, top, bottom]
        m1.drawmeridians(
            np.arange(np.ceil(lonW), np.ceil(lonE), xdelta),
            labels=lonlabs,
            dashes=[1, 1],
            linewidth=0.25,
            color="0.5",
        )
        m1.shadedrelief()
        m1.drawcountries(color="black", linewidth=3)

        vals = data[var].values
        nvals = len(list(set(vals)))

        # check only applies if a built-in colour scheme name is provided
        if isinstance(colours, str):
            if coli > add_cols:
                raise Exception(
                    "Please ensure that coli <= add_cols,"
                    + " otherwise insufficiently many colours"
                    + " are available."
                )
            else:
                colours = sns.color_palette(colours[coli:],
                                            n_colors=nvals + add_cols)

        if not isinstance(colours, str):
            if len(colours) < coli:
                raise Exception("Not enough colours to start from " +
                                str(coli))

        if nvals > len(colours) - coli:
            raise Warning(
                "Insufficiently many colours provided "
                + "once requested colours were removed; "
                + "note that colours will be recycled"
            )

            colours = sns.color_palette(colours[coli:])

        sca = dict()

        for i in range(nvals):
            value = pd.unique(vals)[i]
            loc = vals == value
            sca[i] = m1.scatter(
                _lons[loc],
                _lats[loc],
                latlon=True,
                marker="o",
                s=90,
                zorder=1,
                color=colours[i],
                label=value,
                edgecolors="black",
                linewidth=2,
                alpha=0.9,
            )

        plt.legend(loc=legendloc, title=clabels[k])

    plt.savefig(FigDir + descr + ".png", format="png", dpi=300,
                bbox_inches="tight")
