#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr  4 13:56:59 2021
Plots maps of gridded climate data.
For saving of images to work correctly in in the present form,
a subdirectory "figures" should exist in the parent directory
dof the working directory.
@author: stefaan@csag.uct.ac.za
"""
# %% import

import xarray as xr
import numpy as np
import cartopy.feature as cfeature
import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib as mpl
from cycler import cycler
import seaborn as sns

sns.set_context("notebook")
sns.set_style("darkgrid")

# %% constants

# default colours to use for discrete variable mapping
discrete_colours = ['c', 'g', 'lime', 'gold', 'darkorange', 'purple',
                    'crimson', 'dimgrey', 'black']

hot5 = ['c', 'black', 'red', 'darkorange', 'yellow']
paired9 = sns.color_palette("Paired_r", n_colors=10)[1:9]

# add only one blue
paired9.insert(0, sns.color_palette("Paired", n_colors=10)[1])
figdir = "../figures/"  # edit to change where figures are saved to

# %% plot fns


def cts_gridplot(plottable, bounds, width, height, vlims, clabel="",
                 cols='RdYlBu', extend='both', save=False, log=True):
    """
    Plots a map of a continuous gridded variable using cartopy.

    The map is plotted, and if `save` is set to 'True', it plot is saved
    to file in the directory specified by 'figdir' in this module.
    The PlateCarree projection is used.

    Parameters
    ----------
    plottable : xarray DataArray
        A 2-dimensional (latitude-longitude) variable to be mapped
    bounds : length 4 list or array-like [W, E, S, N]
        The western (W), eastern (E), southern (S) and northern (N) boundaries
        of the domain to be mapped.
    width : float
        Figure width in mm
    height : float
        Figure height in mm
    vlims : tuple, or length 2 list or array
        The minimum and maximum values of the plotted variable to be shown
    clabel : str, optional
         The label of the colourbar. The default is "" (i.e. no label).
    cols : str, or list or array-like of valid colours, optional
        Colourmap name or list of colours to use. The default is 'RdYlBu'.
    extend : str {''}, optional
        Extend the colourbar with triangles beyond the lower ('min'),
        and ('both')/or upper ('max') vlims, to indicate the presence of
        values beyond these limits in the data. For no extention use 'neither'.
        The default is 'both'. See `matplotlib.pyplot.colorbar` for more
        details.
    save : False or str, optional
        Should the plot be saved? If so, provide the file
        name (path) without extension here. The default is False.
    log : bool, optional
        Should a log-scale be used for the colourscale. The default is True.

    Returns
    -------
    None.

    """
    fig = plt.figure(figsize=(width/25, height/25))
    ax = plt.axes(projection=ccrs.PlateCarree())
    ax.coastlines()
    ax.set_extent(bounds, crs=ccrs.PlateCarree())
    gl = ax.gridlines(linewidth=2, color='black',
                      alpha=0.25, draw_labels=True)
    gl.xlabels_top = False
    gl.ylabels_right = False
    ax.add_feature(cfeature.BORDERS)
    if log:
        cnorm = colors.LogNorm(vmin=vlims[0], vmax=vlims[1], clip=True)
    else:
        cnorm = colors.Normalize(vmin=vlims[0], vmax=vlims[1], clip=True)
    pl = plottable.plot(ax=ax, extend=extend, cmap=cols, norm=cnorm,
                        transform=ccrs.PlateCarree(),
                        cbar_kwargs={'shrink': 0.5,
                                     'orientation': 'horizontal',
                                     'pad': 0.07 * (len(clabel) > 0)})
    if clabel != "":
        pl.colorbar.set_label(clabel)
    else:
        pl.colorbar.remove()
    if save:
        plt.savefig(figdir + save, bbox_inches='tight')


def cts_facetplot(plottable, facetvar, ncols, bounds, width, height, vlims,
                  clabel="", cbarorient='horizontal', cols='RdYlBu',
                  extend='both', degs=True, nlabs=6,
                  save=False, log=False):
    """
    Produces an array of 2-D maps of a continuous gridded 3-D variable.

    This function is an extion of cts_gridplot, to produce a grid of maps.
    The input data should have 2 spatial dimensions (latitude-longitude) and
    a third `facetvar` dimension. The facetgrid includes one map for
    each unique value of `facetvar` in `plottable`.
    The PlateCarree projection is used.

    Parameters
    ----------
    plottable : xarray DataArray
        A 3-D xarray DataArray variable, with 2 spatial dimensions
        (latitude-longitude) to map by and a third dimension `facetvar`.
    facetvar : str
        The name of dimension to facet by.
    ncols : int
        Number of columns for the resulting grid.
    bounds : length 4 list or array-like [W, E, S, N]
        The western (W), eastern (E), southern (S) and northern (N) boundaries
        of the domain to be mapped.
    width : float
        Figure width in mm of the grid of maps
    height : float
        Figure height in mm of the grid of maps
    vlims : tuple, or length 2 list or array
        The minimum and maximum values of the plotted variable to be shown.
        This is applied to all facets in the grid.
    clabel : str, optional
         The label of the colourbar. The default is "" (i.e. no label).
    cbarorient : None or {'vertical', 'horizontal'}, optional
        The orientation of the colourbar. The default is 'horizontal'. See
        `matplotlib.pyplot.colorbar` for more details.
    cols : str, or list or array-like of valid colours, optional
        Colourmap name or list of colours to use. The default is 'RdYlBu'.
    extend : str {''}, optional
        Extend the colourbar with triangles beyond the lower ('min'),
        and ('both')/or upper ('max') vlims, to indicate the presence of
        values beyond these limits in the data. For no extention use 'neither'.
        The default is 'both'. See `matplotlib.pyplot.colorbar` for more
        details.
    degs : bool, optional
        Force x- and y-labels to be integer degrees. The default is True.
    nlabs : int, optional
        The approximate number of labels on each axis. The default is 6.
        If `degs` is 'True', the max number is limited by the domain extent.
     save : False or str, optional
         Should the plot be saved? If so, provide the file
         name (path) without extension here. The default is False.
     log : bool, optional
         Should a log-scale be used for the colourscale. The default is False.

    Returns
    -------
    None.

    """

    proj = ccrs.PlateCarree()
    if log:
        cnorm = colors.LogNorm(vmin=vlims[0], vmax=vlims[1], clip=True)
    else:
        cnorm = colors.Normalize(vmin=vlims[0], vmax=vlims[1], clip=True)

    yname, xname = plottable.isel({facetvar: 0}).dims  # non-facetting dims

    nfacets = len(plottable[facetvar])
    nrow = np.ceil(nfacets / ncols)
    padding = max(0.03, 0.12 - 0.03 * nrow)
    facet_aspect = (plottable[xname].size/plottable[yname].size +
                    1 / (2 * ncols) * (cbarorient == "horizontal") *
                    (len(clabel) > 0))

    pl = plottable.plot(extend=extend, cmap=cols, norm=cnorm,
                        col=facetvar, col_wrap=ncols,
                        transform=ccrs.PlateCarree(),
                        add_colorbar=clabel,
                        figsize=(width / 25, height / 25),  # mm => in
                        aspect=facet_aspect,
                        subplot_kws={'projection': proj},
                        cbar_kwargs={'label': clabel,
                                     'shrink': 0.5,
                                     'orientation': cbarorient,
                                     'pad': padding * (len(clabel) > 0)})
    lonW, lonE, latS, latN = bounds
    if degs:
        lonW, latS = np.floor([lonW, latS])  # work with whole degrees
        lonE, latN = np.ceil([lonE, latN])

    # spacing between grid border labels
    gdelta = np.round(max(lonE - lonW, latN - latS) / nlabs, decimals=1 - degs)
    lats = np.arange(latS, latN + gdelta, step=gdelta)
    lons = np.arange(lonW, lonE + gdelta, step=gdelta)

    for ax in pl.axes.flat:
        ax.coastlines()
        ax.set_extent(bounds, crs=ccrs.PlateCarree())
        gl = ax.gridlines(linewidth=2, color='black', xlocs=lons, ylocs=lats,
                          alpha=0.25, draw_labels=True)
        gl.top_labels = False
        gl.right_labels = False
        ax.add_feature(cfeature.BORDERS)
        # ax.set_aspect('equal', 'box')

    if save:
        plt.savefig(figdir + save, bbox_inches='tight')


# %% discrete


def discr_gridplot(ds, var, bounds, x='lon', y='lat', width=200, height=200,
                   cols=discrete_colours, alpha=0.75, degs=True, nlabs=6,
                   legloc='lower left', legtitle='', save=False):
    """
    Plots a map of a discrete gridded variable using cartopy.

    This function is the equivalent of cts_gridplot for a discrete variable.
    It takes in a Dataset and a string name of the variable to plot.
    The map is plotted, and if `save` is set to 'True', it plot is saved
    to file in the directory specified by 'figdir' in this module.
    Note that, since the xarray scatter plot is used and symbol size cannot
    be independently specified, this plot produces reasonable output only if
    the width and height are chosen so that the plot symbols precisely cover
    the size of a gridbox. The PlateCarree projection is used.

    Parameters
    ----------
    ds : xarray Dataset
        Dataset containing the variable to plotted.
    var : str
        Name of the variable in ``ds`` to be plotted.
    bounds :  length 4 list or array-like [W, E, S, N]
        The western (W), eastern (E), southern (S) and northern (N) boundaries
        of the domain to be mapped.
    x : str, optional
        Name of the horizontal axis dimension. The default is 'lon'.
    y : str, optional
        Name of the horizontal axis dimension. The default is 'lat'.
    width : float, optional
        Width of plot in mm. The default is 200.
    height : float, optional
        Height of figure in mm. The default is 200.
    cols : str, or list or array-like of valid colours, optional
        Colourmap name or list of colours to use. The default is:
        ['c', 'g', 'lime', 'gold', 'darkorange', 'purple', 'crimson',
         'dimgrey', 'black'].
    alpha : float in [0, 1], optional
        Transparancy of colours. The default is 0.75.
    degs : bool, optional
        Force x- and y-labels to be integer degrees. The default is True.
    nlabs : int, optional
        The approximate number of labels on each axis. The default is 6.
        If `degs` is 'True', the max number is limited by the domain extent.
    legloc : str in {'best', 'upper right', 'upper left', 'lower left',
                     'lower right', 'right', 'center left', 'center right',
                     'lower center', 'upper center', 'center'}
            pair of floats, or 'int' in [0, 10], optional
        Where the legend is to be located. The default is 'lower left'. See
        `matplotlib.pyplot.scatter` parameter `loc` for more details.
    legtitle : str, optional
        Title of the legend. The default is '' (no title).
    save : False or str, optional
        Should the output map be saved to file? If so, provide the file
        name (path) without extension here. The default is False.

    Returns
    -------
    None.

    """

    # change the default colours matplotlib cycles through
    mpl.rcParams['axes.prop_cycle'] = cycler(color=cols)

    # setup the cartopy map
    fig = plt.figure(figsize=(width / 25, height / 25))
    ax = plt.axes(projection=ccrs.PlateCarree())
    ax.coastlines()
    ax.set_extent(bounds, crs=ccrs.PlateCarree())
    lonW, lonE, latS, latN = bounds
    if degs:
        lonW, latS = np.floor([lonW, latS])  # work with whole degrees
        lonE, latN = np.ceil([lonE, latN])
    gdelta = np.round(max(lonE-lonW, latN-latS)/nlabs, decimals=1-degs)
    lats = np.arange(latS, latN+gdelta, step=gdelta)
    lons = np.arange(lonW, lonE+gdelta, step=gdelta)

    # grid lines
    gl = ax.gridlines(linewidth=2, color='black', xlocs=lons, ylocs=lats,
                      alpha=0.25, draw_labels=True)
    gl.top_labels = False
    gl.right_labels = False
    ax.add_feature(cfeature.BORDERS)

    xr.plot.scatter(ds, x, y, hue=var, ax=ax, add_guide=False,
                    transform=ccrs.PlateCarree(), marker='s', alpha=alpha)
    plt.legend(loc=legloc, title=legtitle)

    if save:
        plt.savefig(figdir + save, bbox_inches='tight', dpi=300)
