#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Mar  3 16:39:06 2021

@author: stefaan
"""

# %% imports

import cartopy.crs as ccrs
import numpy as np
import matplotlib.pyplot as plt
import cartopy.io.img_tiles as cimgt
import pandas as pd

# %% constants

FigDir = "../figures/"

# %% functions


def annotated_map(labeldicts, linedicts, msize=12, fsize=14, savemeta="SWSA",
                  lonW=16, lonE=22, latN=-27.5, latS=-35.2, zoom=8,
                  figwidth=350, figheight=500):
    """
    Draws a labelled Stamen terrain map

    For more information about the underlying map see: http://stamen.com/ and
    https://www.openstreetmap.org/ (OSM)

    The zoom-level and domain (a lat-lon gridbox) can also be specified.

    The domain defaults to OSM-zoom level 8 and the default domain is the
    region encompassing the South African Winter Rainfall Zone

    Labels are specified in a list of dictionaries (`labeldicts`),
    with entries (keys) for:
        * the label text (name),
        * latitude (lat),
        * longitude (lon),
        * and a boolean specifying whether markers should be drawn (marker);
    background boxes are drawn for marked points only.
    Optionally additional entries can be added to dictionaries to specify:
        * the text label offsets from marked points (xoffset, yoffset),
        * the alignment of the text (xalign, yalign),
        * the label orientation (angle)
        * and the text (& marker) colour (colour).
    Lines are drawn on the map with characteristics specified in
    a list of dictionaries (`linedicts`) with entries (keys) for
    [only co-ordinates are required]:
        * the co-ordinates of the start (lon0, lat0) and end (lon1, lat1) pnts
        * the linestyle to be used (ls; optional)
        * the linewidth to apply (lw; optional)
        * the colour for the line (lc; optional)
    Figure dimensions are in mm.


    Parameters
    ----------
    labeldicts : dictionary
        Contains information regarding what labels should be drawn where.
        See detailed description above
    linedicts : dictionary
        As labeldicts but for lines
    msize : int, optional
        marker size for points plotted. The default is 12.
    fsize : int, optional
        Font size for labels. The default is 14.
    savemeta : str, optional
        Description to use for saving the map. The default is "SWSA".
    lonW : float, optional
        Western-edge longitude. The default is 16.
    lonE : float, optional
        Eastern-edge longitude. The default is 22.
    latN : float, optional
        Northern-edge latitude. The default is -27.5.
    latS : float, optional
        Southern-edge latitude. The default is -35.2.
    zoom : int, optional
        OSM zoom level to use. The default is 8.
    figwidth : float, optional
        Map width. The default is 350.
    figheight : float, optional
        Map height. The default is 500.

    Raises
    ------
    Exception
        If the required dictionary keys are not provided for labeldicts or
        linedicts

    Returns
    -------
    None.

    """
    stamen_terrain = cimgt.Stamen('terrain-background')

    # convert mm => in
    fig = plt.figure(figsize=(figwidth/25.4, figheight/25.4))
    ax = fig.add_subplot(1, 1, 1, projection=stamen_terrain.crs)

    ax.set_extent([lonW, lonE, latS, latN], crs=ccrs.Geodetic())
    ax.add_image(stamen_terrain, zoom)

    nlab = len(labeldicts)
    nlin = len(linedicts)

    requiredkeys = ['name', 'lat', 'lon', 'marker']

    # allowed label orientation specified as string
    stringorient = ['vertical', 'horizontal']

    # make labels
    for i in np.arange(nlab):
        di = labeldicts[i]

        if emptykeylist(di, requiredkeys):
            raise Exception("Please ensure that all dictionaries contain" +
                            " valid entries (values) for the following" +
                            " required keys: " + ", ".join(requiredkeys))

        # defaults for label properties if not specified

        if emptykey(di, 'colour'):
            di['colour'] = 'black'

        if emptykey(di, 'angle'):
            di['angle'] = 0

        # only specified strings and float angles are allowed as input
        elif di['angle'] not in stringorient:
            di['angle'] = np.float(di['angle'])

        # default label horizontal offset of 0
        if emptykey(di, 'xoffset'):
            di['xoffset'] = 0

        # vertical label offset of 0.08 degrees only if a marker is used.
        if emptykey(di, 'yoffset'):
            di['yoffset'] = 0.08*di['marker']

        if emptykey(di, 'xalign'):
            di['xalign'] = 'center'

        # meridional alignment:
        # if y offset is above: align bottom
        # if y offset is 0: align center
        # if y offset is below: align top
        if emptykey(di, 'yalign'):
            di['yalign'] = ('center' * np.int(di['yoffset'] == 0) +
                            'bottom' * np.int(di['yoffset'] > 0) +
                            'top' * np.int(di['yoffset'] < 0))

        if di['marker']:
            ax.plot(di['lon'], di['lat'], marker='o', color=di['colour'],
                    markersize=msize, alpha=0.7, transform=ccrs.Geodetic())

            ax.text(x=di['lon']+di['xoffset'], y=di['lat']+di['yoffset'],
                    s=di['name'], fontsize=fsize, transform=ccrs.Geodetic(),
                    verticalalignment=di['yalign'],
                    horizontalalignment=di['xalign'],
                    rotation=di['angle'], c=di['colour'],
                    bbox=dict(facecolor='white', alpha=0.7, boxstyle='round'))
        else:
            ax.text(x=di['lon'], y=di['lat'], s=di['name'], fontsize=fsize,
                    transform=ccrs.Geodetic(),
                    verticalalignment=di['yalign'],
                    horizontalalignment=di['xalign'],
                    rotation=di['angle'], c=di['colour'],
                    bbox=dict(facecolor='white', alpha=0.7, boxstyle='round'))

    requiredkeys = ['lon0', 'lat0', 'lon1', 'lat1']

    for i in np.arange(nlin):
        di = linedicts[i]

        if emptykeylist(di, requiredkeys):
            raise Exception("Please ensure that all dictionaries contain" +
                            " valid entries (values) for the following" +
                            " required keys: " + ", ".join(requiredkeys))

        # defaults:
        if emptykey(di, 'lc'):
            di['lc'] = 'white'

        if emptykey(di, 'ls'):
            di['ls'] = '--'

        if emptykey(di, 'lw'):
            di['lw'] = 6

        plt.plot([di['lon0'], di['lon1']], [di['lat0'], di['lat1']],
                 linewidth=di['lw'], color=di['lc'], linestyle=di['ls'],
                 transform=ccrs.Geodetic())

    plt.savefig(FigDir+savemeta+"_topogmap.png",
                bbox_inches='tight')


# %% Helper functions

def emptykeylist(di, keylist):
    return (np.sum([emptykey(di, key) for key in keylist]))


def emptykey(di, key):
    return ((key not in di) or pd.isnull(di[key]) or di[key] == '')


def df2dictlist(df):
    nd = len(df)   # number of dictionaries
    dictlist = []
    for i in np.arange(nd):
        di = dict(df.iloc[i])
        dictlist.append({k: v for k, v in di.items() if not pd.isnull(v)})
    return dictlist


def preserve_newline(df, cols):
    nr = len(df)
    for col in cols:
        df[col] = [df[col][i].replace('\\n', '\n') for i in np.arange(nr)]
    return df
