#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Nov  3 23:18:20 2018

@author: stefaan
"""
# %%

import numpy as np
from sklearn.cluster import FeatureAgglomeration
import matplotlib.pyplot as plt
import fn.xrclim_fns as clm
import fn.basemap_clusterplots as bmc
import fn.basemap_plots as bmp

# %%

###===================================================
###
### Constant definitions
###
###===================================================

# path name components

figdir = "../figures/"
datadir = "../data/"

# spatial bounds

latN = -28.5
latS = -35.2
lonW = 16.3
lonE = 21.2

# dummy initial labels where not otherwise specified

labs = ["A", "B", "D", "F", "G", "H", "J", "K", "L", "M", "N", "P", "Q"]


labsW = ["E", "N", "S", "CPT", "SW", "SE", "W", "AA", "BB", "DD"]
labsV = ["NW", "CPT", "SW", "N", "NE", "S", "SE", "E", "BB", "DD"]

labsW5yr = ["E", "S", "N"]
labsV5yr = ["transition", "SW", "NE"]

windlabs = [
    "Core",
    "Periphery",
    "Escarpment",
    "All-year",
    "SCoast",
    "E",
    "inner",
    "CPT",
    "interior",
    "DD",
]

# import colour sets

# Qualitative colour palettes

# Hex lists of colours

# Tol colours:

# blue red green yellow cyan purple grey black
Tol_bright = [
    "#4477AA",
    "#EE6677",
    "#228833",
    "#CCBB44",
    "#66CCEE",
    "#AA3377",
    "#BBBBBB",
    "#000000",
]

# my alternatives, with a different grey, at the end  (no. 4) and with
# black first

# blue red green yellow cyan purple black grey
Tol_b2 = [
    "#4477AA",
    "#EE6677",
    "#228833",
    "#CCBB44",
    "#66CCEE",
    "#AA3377",
    "#000000",
    "#555555",
]

# blue red green grey yellow cyan purple black
Tol_b3 = [
    "#4477AA",
    "#EE6677",
    "#228833",
    "#555555",
    "#CCBB44",
    "#66CCEE",
    "#AA3377",
    "#000000",
]

Tol_b1 = [
    "#000000",
    "#EE6677",
    "#CCBB44",
    "#66CCEE",
    "#228833",
    "#555555",
    "#4477AA",
    "#AA3377",
]  # black red yellow cyan green grey blue purple

# high contrast

Tol_hc = ["#004488", "#DDAA33", "#BB5566", "#000000"]  # blue yellow red black

# orange blue cyan magenta red teal grey black
Tol_vibrant = [
    "#EE7733",
    "#0077BB",
    "#33BBEE",
    "#EE3377",
    "#CC3311",
    "#009988",
    "#BBBBBB",
    "#000000",
]

# rose indigo sand green cyan wine teal olive purple pale_grey black
Tol_muted = [
    "#CC6677",
    "#332288",
    "#DDCC77",
    "#117733",
    "#88CCEE",
    "#882255",
    "#44AA99",
    "#999933",
    "#AA4499",
    "#DDDDDD",
    "#000000",
]
# orange red rose purple wine indigo blue cyan teal green olive sand black
# pale_grey
my15 = [
    "#EE7733",
    "#CC3311",
    "#CC6677",
    "#AA4499",
    "#882255",
    "#332288",
    "#0077BB",
    "#88CCEE",
    "#44AA99",
    "#117733",
    "#999933",
    "#DDCC77",
    "#000000",
    "#DDDDDD",
]

# my adaptation of Tol_muted for when first colour is to be reserved for the
# ocean, 2nd colour for single entity and other colours come in 2s; last
# colour for NA

# cyan black indigo wine sand olive green teal pale_grey
Tol_seafirst_paired = [
    "#88CCEE",
    "#000000",
    "#332288",
    "#882255",
    "#DDCC77",
    "#999933",
    "#117733",
    "#44AA99",
    "#DDDDDD",
]

# light_blue orange light_yellow pink light_cyan mint pear olive pale_grey
# black
Tol_light = [
    "#77AADD",
    "#EE8866",
    "#EEDD88",
    "#FFAABB",
    "#99DDFF",
    "#44BB99",
    "#BBCC33",
    "#AAAA00",
    "#DDDDDD",
    "#000000",
]

# Wong 2011 (nature methods) colours:

# Black, Orange, Sky_Blue, Bluish_Green, Yellow, Blue, Vermillion,
# Reddish purple
Wong = [
    "#000000",
    "#E69D00",
    "#56B4E9",
    "#009E74",
    "#F0E442",
    "#0072B2",
    "#D55C00",
    "#CC79A7",
]

# IBM

IBM = ["#648FFF", "#785EF0", "#DC267F", "FE6100", "FFB000"]

# %% clustering function


def clust_stations(data, n_clusters):
    """
    Perform hierarchical Ward's Clustering with a set number of clusters.

    Parameters
    ----------
    data : 2-D array or xarray DataArray
        The variables to cluster on; first dimension for the different
        variables to cluster by, second for the objects to cluster.
    n_clusters : float
        Number of clusters to find

    For further details see `sklearn.cluster.FeatureAgglomeration`.

    """
    return FeatureAgglomeration(n_clusters, linkage="ward").fit(data)


# %%  Cluster by seasonality (normalised monthly, with December discarded)


def climclust_plots(
    datavar,
    lon,
    lat,
    meta,
    clustmin=5,
    clustmax=8,
    clustopt=8,
    bounds=[lonW, lonE, latS, latN],
    legloc="lower left",
    var="rain",
    labs=labs,
    colours=Tol_bright,
):
    """
    Cluster station data by proportion of variable occurring in each month.

    Plots spatial map of the clusters stations are categorised into and
    line plots of the mean normalised climatology in each cluster.
    The input meta data file is updated in place with a column added for
    the climatology cluster obtained. Written output is for the predetermined
    "optimal" number of clusters, plots are produced for each number of
    clusters between clustmin and clustmax.

    Parameters
    ----------
    datavar : xarray DataArray
        The variable to cluster by. Monthly totals of this variable should be
        meaningful.
    lons : list of float
        Longitude of each station (or equivalent point location) to be mapped.
        Must be the same length as the "station" (or equivalent) dimension
        in `datavar`.
    lats : list of float
        Latitude of each station (or equivalent point location) to be mapped.
        Must be the same length as the "station" (or equivalent) dimension
        in `datavar`.
    meta : pandas DataFrame
        Dataframe containing metadata about the stations, to which a new
        column should be added recording the cluster each station was assigned
        to.
    clustmin : int, optional
        Perform clustering from this value to `clustmax`. The default is 5.
    clustmax : int, optional
        The default is 8.
    clustopt : int, optional
        Number of clusters to write cluster values to `meta` for.
        The default is 8.
   bounds : length-4 list or array-like of floats
       A list = [lonW, lonE, latS, latN], where lonW and lonE are
       the longitudes on the western and eastern edges, respectively of the
       rectangular map, and latS and latN the latitudes at its respective
       southern and northern edges.
    legloc : str in {'best', 'upper right', 'upper left', 'lower left',
                       'lower right', 'right', 'center left', 'center right',
                       'lower center', 'upper center', 'center'}; or
            pair of floats; or 'int' in [0, 10], optional
        Where the legend is to be located. The default is 'lower left'. See
        `matplotlib.pyplot.scatter` parameter `loc` for more details.
        The default is "lower left".
    var : str, optional
        The name to use to refer to the clustering variable by in plots and
        when saving to disk. The default is 'rain'.
    labs :  list or array of str
        Names of the clusters to use in the legend.
        The default is dummy list `station_clusters.labs`.
    colours : list, or str optional
        list or list-like of colour names or string defining a built-in
        seaborn colourmap. The default is Tol_bright [defined in this script].

    Returns
    -------
    None.

    """

    clim, nclim, nclim11 = clm.climatology(datavar)

    # Climatology clustering

    descr = var + "_climclust"

    m1, fig = bmp.station_basemap(bounds=bounds, width=120, height=150)

    for nclust in np.arange(clustmin, clustmax + 1):
        ward_clim = clust_stations(nclim11, nclust)

        bmc.cluster_map(
            ward_clim,
            nclust,
            m1,
            colours=colours,
            labs=labs,
            lons=lon,
            lats=lat,
            desc=descr,
            legloc=legloc,
        )

    if clustmin == clustmax:  # reuse existing wardclim

        bmc.seasonality_plot(
            clusternums=ward_clim.labels_,
            climdata=nclim,
            station_num=datavar.station.argsort(),
            desc=descr,
            regnames=labs,
            normclim=True,
            colours=colours,
            varname=var,
            stationdim="station",
        )

    else:
        for nclust in np.arange(clustmin, clustmax + 1):
            ward_clim = clust_stations(nclim11, nclust)

            bmc.seasonality_plot(
                clusternums=ward_clim.labels_,
                climdata=nclim,
                station_num=datavar.station.argsort(),
                desc=descr,
                regnames=labs,
                normclim=True,
                varname=var,
                stationdim="station",
                colours=colours,
            )

    nclust = clustopt
    ward_clim = clust_stations(nclim11, nclust)
    clustS = ward_clim.labels_
    meta["seas"] = [labs[j] for j in clustS]

    # changes made in place, so not necessary to return


# %% Standardised interannual variability clustering (by year as dim) ###


def varclust_plots(
    datavar,
    lon,
    lat,
    meta,
    clustmin=5,
    clustmax=8,
    optV=8,
    optW=7,
    colours=Tol_b3,
    bounds=[lonW, lonE, latS, latN],
    legloc="lower left",
    var="rain",
    labsV=labsV,
    labsW=labsW,
    drop2=False,
):
    """
    As for `climclust_plots()` but clustering by temporal variability.

    Clustering is performed both on annual totals and on "winter" (AMJJAS)
    totals. Standardised rainfall anomalies are used as the clustering
    variables.

    Parameters
    ----------
    datavar : xarray DataArray
        The variable whose annual and seasonal totals to cluster by.
    lons : list of float
        Longitude of each station (or equivalent point location) to be mapped.
        Must be the same length as the "station" (or equivalent) dimension
        in `datavar`.
    lats : list of float
        Latitude of each station (or equivalent point location) to be mapped.
        Must be the same length as the "station" (or equivalent) dimension
        in `datavar`.
    meta : pandas DataFrame
        Dataframe containing metadata about the stations, to which a new
        column should be added recording the cluster each station was assigned
        to.
    clustmin : int, optional
        Perform clustering from this value to `clustmax`. The default is 5.
    clustmax : int, optional
        The default is 8.
    optV : int, optional
        Number of clusters to write cluster values to `meta` for, for annual
        totals. The default is 8.
    optW : int, optional
        As with optV, but for AMJJAS. The default is 7.
    colours : TYPE, optional
        DESCRIPTION. The default is Tol_b3.
    bounds : length-4 list or array-like of floats
        A list = [lonW, lonE, latS, latN], where lonW and lonE are
        the longitudes on the western and eastern edges, respectively of the
        rectangular map, and latS and latN the latitudes at its respective
        southern and northern edges.
     legloc : str in {'best', 'upper right', 'upper left', 'lower left',
                        'lower right', 'right', 'center left', 'center right',
                        'lower center', 'upper center', 'center'}; or
             pair of floats; or 'int' in [0, 10], optional
         Where the legend is to be located. The default is 'lower left'. See
         `matplotlib.pyplot.scatter` parameter `loc` for more details.
         The default is "lower left".
     var : str, optional
         The name to use to refer to the clustering variable by in plots and
         when saving to disk. The default is 'rain'.
    labsV : list or array of str, optional
        Names of the clusters to use in the legend for the annual plot.
        The default is `station_clusters.labsV`.
    labsW : list or array of str, optional
        As for labsV, but for AMMJAS. The default is `station_clusters.labsW`.
    drop2 : bool, optional
        Whether to do a correction for the linear dependence introduced by
        standardisation, but removing redundant data. The default is False.

    Returns
    -------
    None.

    """

    svar = clm.std_ia(datavar, drop2=drop2)
    svarw = clm.std_ia_ew(datavar, drop2=drop2)  # extended winter only

    svar = svar.where(~svar.isnull(), other=0)  # climatology for missing
    svarw = svarw.where(~svarw.isnull(), other=0)

    m1, fig = bmp.station_basemap(bounds=bounds, width=100, height=120)

    descr = var + "_varclust_"

    for nclust in np.arange(clustmin, clustmax + 1):
        ward_var = clust_stations(svar, nclust)
        ward_winter = clust_stations(svarw, nclust)

        bmc.cluster_map(
            ward_var,
            nclust,
            m1,
            colours=colours,
            labs=labsV,
            lons=lon,
            lats=lat,
            desc=descr + "annual",
            legloc=legloc,
        )

        bmc.cluster_map(
            ward_winter,
            nclust,
            m1,
            colours=colours,
            labs=labsW,
            lons=lon,
            lats=lat,
            desc=descr + "AMJJAS",
            legloc=legloc,
        )

    for nclust in np.arange(clustmin, clustmax + 1):
        bmc.variability_plot(
            svar,
            ward_var.labels_,
            nclust,
            labsV,
            descr + "annual",
            "annual",
            colours=colours,
        )
        bmc.variability_plot(
            datavar,
            ward_var.labels_,
            nclust,
            labsV,
            descr + "annual",
            "annual",
            std=False,
            colours=colours,
        )
        bmc.variability_plot(
            svarw,
            ward_winter.labels_,
            nclust,
            labsW,
            descr + "AMJJAS",
            "AMJJAS",
            colours=colours,
        )
        varAMJJAS = datavar.where(clm.AMJJAS(datavar["time.month"]),
                                  drop=True)
        bmc.variability_plot(
            varAMJJAS,
            ward_winter.labels_,
            nclust,
            labsW,
            descr + "AMJJAS",
            "AMJJAS",
            std=False,
            colours=colours,
        )

    nclust = optW

    ward_winter = clust_stations(svarw, nclust)
    clustW = ward_winter.labels_
    meta["wvar"] = [labsW[j] for j in clustW]

    nclust = optV

    ward_var = clust_stations(svar, nclust)
    clustV = ward_var.labels_
    meta["avar"] = [labsV[j] for j in clustV]  # changes made in place, so
    # not necessary to return


def blockvarclust_plots(
    datavar,
    lon,
    lat,
    meta,
    nyr=5,
    clustmin=5,
    clustmax=8,
    optV=8,
    optW=7,
    colours=Tol_bright,
    var="rain",
    bounds=[lonW, lonE, latS, latN],
    legloc="lower left",
    labsV=labsV5yr,
    labsW=labsW5yr,
):
    """
    Equivalent of `varclust_plots()` for non-overlapping `nyr` aggregates.

    Parameters
    ----------
    See `varclust_plots()`. One additional parameter:
    nyr : int, optional
        Number of years to aggregate. The default is 5.

    Returns
    -------
    None.

    """
    yr0 = datavar["time.year"].min()
    datavar["year"] = np.floor((datavar["time.year"] - yr0) / nyr)

    svar = clm.std_ia(datavar, groupvar="year", yrvar="year")
    svarw = clm.std_ia_ew(datavar, groupvar="year", yrvar="year")

    svar = svar.where(~svar.isnull(), other=0)  # climatology for missing
    svarw = svarw.where(~svarw.isnull(), other=0)

    m1, fig = bmp.station_basemap(bounds=bounds, width=120, height=180)

    descr = var + "_varclust"

    for nclust in np.arange(clustmin, clustmax + 1):
        ward_var = clust_stations(svar, nclust)
        ward_winter = clust_stations(svarw, nclust)

        bmc.cluster_map(
            ward_var,
            nclust,
            m1,
            colours=colours,
            labs=labsV,
            lons=lon,
            lats=lat,
            legloc=legloc,
            desc=descr + "annual_" + str(nyr) + "yr_block",
        )

        bmc.cluster_map(
            ward_winter,
            nclust,
            m1,
            colours=colours,
            labs=labsW,
            lons=lon,
            lats=lat,
            legloc=legloc,
            desc=descr + "AMJJAS_" + str(nyr) + "yr_block",
        )

    # year roughly in middle of each block
    svar["year"] = yr0 + svar.year * nyr + np.floor(nyr / 2)
    svarw["year"] = yr0 + svarw.year * nyr + np.floor(nyr / 2)

    for nclust in np.arange(clustmin, clustmax + 1):
        bmc.variability_plot(
            svar,
            ward_var.labels_,
            nclust,
            labsV,
            descr + "annual_" + str(nyr) + "block",
            "annual",
            colours=colours,
        )

        bmc.variability_plot(
            svarw,
            ward_winter.labels_,
            nclust,
            labsW,
            descr + "AMJJAS_" + str(nyr) + "block",
            "AMJJAS",
            colours=colours,
        )

    nclust = optW

    ward_winter = clust_stations(svarw, nclust)
    clustW = ward_winter.labels_
    meta["wvar_block" + str(nyr)] = [labsW[j] for j in clustW]

    nclust = optV

    ward_var = clust_stations(svar, nclust)
    clustV = ward_var.labels_
    meta["avar_block" + str(nyr)] = [labsV[j] for j in clustV]
    # changes made in place, so
    # not necessary to return


# %% cluster by contribution (wind directions)


def wdirn_cluster(
    datavar,
    meta,
    clustmin=5,
    clustmax=8,
    opt=8,
    contrib_dim="CWT",
    stationdim="station",
    addmu=True,
    dropdirn="WF",
    labs=windlabs,
    descr="by_CWT",
    ncol=2,
    colours=Tol_b1,
    bounds=[16.8, 21.5, -34.9, -28.8],
    legloc="upper right",
    widthmap=135,
    heightmap=160,
    widthbar=80,
    heightbar=25,
):
    """
    Performs clustering on CWT (wind direction bins) rainfall contributions.

    Plots a scatter plot on a Basemap indicating which cluster each station
    is assigned to. A bar plot indicating the climatological spatial mean
    contribution of each CWT in each cluster to the annual mean rainfall.

    Parameters
    ----------
    datavar : 2-D xarray DataArray
        Contribution by `contrib_dim` for each station in `stationdim`.
    meta : pandas DataFrame
        Dataframe containing metadata about the stations, to which a new
        column should be added recording the cluster each station was assigned
        to.
    clustmin : int, optional
        Perform clustering from this value to `clustmax`. The default is 5.
    clustmax : int, optional
    opt : int, optional
        Number of clusters to write cluster values to `meta` for.
        The default is 8.
    contrib_dim : str, optional
        The dimension in `datavar` whose unique instances are to be used as
        clustering variables. The default is 'CWT'.
    stationdim : str, optional
        The name of the dimension in `datavar` containing the entities to
        cluster. The default is 'station'.
    addmu : bool, optional
        Should a facet be added to the barplot facetgrid showing the overall
        mean contributions across all CWTs? The default is True.
    dropdirn : str, optional
        The instance of `contrib_dim` to be ignored as variable when
        clustering. The default is 'WF'.
    labs : list of str, optional
        Names of clusters for legends in map and times series plots.
        The default is `station_clusters:windlabs`.
    descr : str, optional
        Description of the clustering performed. The default is 'by_CWT'.
    ncol : int, optional
        Number of columns in the facetgrid by `contrib_dim` in the barplots.
        The default is 2.
    colours : list, or str, optional
        list or list-like of colour names or string defining a built-in
        seaborn colourmap.The default is `station_clusters.Tol_b1`.
    bounds : length-4 list or array-like of floats
        A list = [lonW, lonE, latS, latN], where lonW and lonE are
        the longitudes on the western and eastern edges, respectively of the
        rectangular map, and latS and latN the latitudes at its respective
        southern and northern edges.The default is [16.8, 21.5, -34.9, -28.8].
    legloc : str in {'best', 'upper right', 'upper left', 'lower left',
                   'lower right', 'right', 'center left', 'center right',
                   'lower center', 'upper center', 'center'}; or
        pair of floats; or 'int' in [0, 10], optional
        Where the legend is to be located. The default is 'lower left'. See
        `matplotlib.pyplot.scatter` parameter `loc` for more details.
        The default is "upper right".
    widthmap : float, optional
        Width in mm for the Basemap. The default is 135.
    heightmap : float, optional
        Height in mm for the Basemap. The default is 160.
    widthbar : float, optional
        Width of each barplot facet panel in mm. The default is 80.
    heightbar : float, optional
       Height of each barplot facet panel in mm . The default is 25.

    Returns
    -------
    None.

    """
    # because of normalisation not all dirns are orthog.; 1 to be dropped

    m1, fig = bmp.station_basemap(bounds=bounds, width=widthmap,
                                  height=heightmap)

    descr = "contribclust_" + descr

    seldirns = [x for x in datavar[contrib_dim].values if x != dropdirn]

    for nclust in np.arange(clustmin, clustmax + 1):
        ward_contrib = clust_stations(datavar.sel({contrib_dim: seldirns}),
                                      nclust)

        bmc.cluster_map(
            ward_contrib,
            nclust,
            m1,
            colours=colours,
            labs=labs,
            lons=meta.loc[datavar[stationdim], "lon"],
            lats=meta.loc[datavar[stationdim], "lat"],
            desc=descr,
            legloc=legloc,
            size=80,
        )

    for nclust in np.arange(clustmin, clustmax + 1):
        ward_contrib = clust_stations(datavar.sel({contrib_dim: seldirns}),
                                      nclust)
        # number of rows after mean is (optionally) added
        nr = np.ceil((nclust + 1 * addmu) / ncol)
        bmc.contrib_plot(
            datavar,
            ward_contrib.labels_,
            nclust,
            labs,
            descr,
            contribdim=contrib_dim,
            ncol=ncol,
            addmu=addmu,
            width=widthbar * (1 + 0.8 * (ncol - 1)),
            height=heightbar * (1 + 0.75 * (nr - 1)),
            colours=colours,
        )

    # write cluster labels from "optimum" number to meta file
    nclust = opt

    ward_contrib = clust_stations(datavar.sel({contrib_dim: seldirns}),
                                  nclust)
    clust = ward_contrib.labels_
    meta["CWT_contrib"] = [labs[j] for j in clust]


# %% Fourier co-efficient clusters


def fft_cluster(
    data,
    meta,
    descr,
    freqs=None,
    clustmin=5,
    clustmax=8,
    opt=8,
    colours=Tol_vibrant,
    labs=labsV,
    stationdim="station",
    bounds=[lonW, lonE, latS, latN],
    legloc="lower left",
):
    """
    Perform clustering on Fourier decomposition co-efficients of rainfall.

    Parameters
    ----------
    data : pandas DataFrame
        Fourier transform output to cluster by.
        A dataframe with the structure produced by metR::FitWave. Columns
        should include "r2" for normalised amplitude, "k" for frequency,
        "phase" for the phase shift, "lon" and "lat" for the longitude and
        latitude of each station, and a `stationdim` column containing
        station names.
    meta : pandas DataFrame
        Dataframe containing metadata about the stations, to which a new
        column should be added recording the cluster each station was assigned
        to.
    descr : str
        Description of the clustering performed.
    freqs : None, or array, optional
        The frequencies to consider (in units of /series length). If None,
        use all. The default is None.
    clustmin : int, optional
        Perform clustering from this value to `clustmax`. The default is 5.
    clustmax : int, optional
    opt : int, optional
        Number of clusters to write cluster values to `meta` for.
        The default is 8.
    colours : list, or str, optional
        list or list-like of colour names or string defining a built-in
        seaborn colourmap. The default is `station_clusters.Tol_vibrant`.
    labs : list of str, optional
        Names of clusters for legends in map and times series plots.
        The default is `station_clusters.labsV`.
    stationdim : str, optional
        The name of the dimension in `datavar` containing the entities to
        cluster. The default is 'station'.
    bounds : length-4 list or array-like of floats, optional
        A list = [lonW, lonE, latS, latN], where lonW and lonE are
        the longitudes on the western and eastern edges, respectively of the
        rectangular map, and latS and latN the latitudes at its respective
        southern and northern edges. The default is [16.3, 21.2, -35.2, -28.5].
    legloc : str in {'best', 'upper right', 'upper left', 'lower left',
                        'lower right', 'right', 'center left', 'center right',
                        'lower center', 'upper center', 'center'}; or
             pair of floats; or 'int' in [0, 10], optional
         Where the map legend is to be located. The default is 'lower left'.
         See `matplotlib.pyplot.scatter` parameter `loc` for more details.
         The default is "lower left".

    Returns
    -------
    None.

    """
    # phase and amplitude not good for clustering;
    # *phase irrelevant when ampl << 1
    # *angles very difficult to compare due to periodic nature
    # hence find real and imaginary part of the co-efficients
    data["a_n"] = data["r2"] * np.cos(data["phase"])
    data["b_n"] = data["r2"] * np.sin(data["phase"])

    if freqs is None:
        freqs = np.unique(data["k"])

    # subset data and get into wide format for clustering
    data = data[data["k"].isin(freqs)].pivot(
        index=stationdim, columns="k", values=["a_n", "b_n"]
    )

    m1, fig = bmp.station_basemap(bounds=bounds, width=120, height=180)

    for nclust in np.arange(clustmin, clustmax + 1):
        ward_var = clust_stations(data.T, nclust)
        bmc.cluster_map(
            ward_var,
            nclust,
            m1,
            colours=colours,
            labs=labs,
            lons=meta.loc[data.index, "lon"],
            lats=meta.loc[data.index, "lat"],
            legloc=legloc,
            desc="fftclust_" + descr,
        )

    nclust = opt
    ward_var = clust_stations(data.T, nclust)
    clust = ward_var.labels_
    meta["fft"] = [labs[j] for j in clust]
