#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Aug 10 19:41:15 2020

@author: stefaan
"""
# %% Imports

import pandas as pd
import xarray as xr
import numpy as np
from datetime import timedelta
import fn.xrclim_fns as xrc

# %% basic wind speed and dirn calc funcs


def calc_wdirn(u, v):
    """
    Given spatio-temporal horizontal wind components, returns wind direction.

    Wind direction is returned in the standard bearing form in [0, 360]:
    i.e. the angle measured clockwise from north to the direction the wind
    is blowing from.

    Parameters
    ----------
    u : xarray DataArray
        Zonal wind component (can have space and time dimensions)
    v : xarray DataArray
        Meridional wind component (must have the same strucuture as v)

    Returns
    -------
    theta : xarray DataArray
        Horizontal wind direction as bearing (angle clockwise from North)

    """
    theta = (90 - np.rad2deg(np.arctan2(v, u))) + 180
    theta = theta.where(theta < 360, other=theta - 360)
    return theta


def calc_mean_wdirn(u, v):
    """
    Direction of the spatial mean wind vector.

    The parameters `u` and `v` are xarray DataArrays of at least 2-D and
    must have "latitude" and "longitude" as their spatial dimensions.

    Wind direction is returned in the standard bearing form in [0, 360]:
    i.e. the angle measured clockwise from north to the direction the wind
    is blowing from.

    Parameters
    ----------
    u : xarray DataArray
        Zonal wind component (can have space and time dimensions)
    v : xarray DataArray
        Meridional wind component (must have the same strucuture as v)

    Returns
    -------
    thetamu : xarray DataArray
        Horizontal wind direction of the spatial mean wind vector as
        bearing (angle clockwise from North)

    """
    umu = u.mean(dim=["latitude", "longitude"])
    vmu = v.mean(dim=["latitude", "longitude"])
    thetamu = (90 - np.rad2deg(np.arctan2(vmu, umu))) + 180
    thetamu = thetamu.where(thetamu < 360, other=thetamu - 360)
    return thetamu


def calc_mean_wspd(u, v):
    """
    Magnitude (speed) of the spatial mean wind vector.

    The parameters `u` and `v` are xarray DataArrays of at least 2-D and
    must have "latitude" and "longitude" as their spatial dimensions.

    Parameters
    ----------
    u : xarray DataArray
        Zonal wind component (can have space and time dimensions)
    v : xarray DataArray
        Meridional wind component (must have the same strucuture as v)

    Returns
    -------
    speedmu :  xarray DataArray
        Horizontal mean wind speed
    """
    umu = u.mean(dim=["latitude", "longitude"])
    vmu = v.mean(dim=["latitude", "longitude"])
    speedmu = np.sqrt(umu ** 2 + vmu ** 2)
    return speedmu


# %% dataset prep


def prep_wind(u, v, IF_thres=1.5, WF_thres=2.5, uname="ua", vname="va"):
    """
    Construct a single xarray Dataset with wind properties as variables.

    Takes two xarray Datasets for the zonal and meridional components of
    horizontal wind speed and constructs a single dataset containing both
    components, as well as the wind speed ["speed"] and direction ["wdirn"]
    at each point, the spatial mean wind speed ["ave_speed"],
    the speed of the vector mean flow ["muflow_speed"],
    and the direction of the mean wind flow vector ["muflow_dirn"].

    For use in assigning CWTs, a further variable "ls_dirn" for the
    large-scale flow direction is also added. To have contiguous sectors for
    each wind direction bin, this variable is mapped to [-22.5, 337.5].
    Time steps with incoherent flow are artificially assigned a ls_dirn of 400;
    those with weak flow are assigned 350.

    Parameters
    ----------
    u : xarray Dataset
        Must contain a spatio-temporal zonal wind component variable named
        `uname` [defualt 'ua'].
    v : xarray Dataset
        Must contain a spatio-temporal meridional wind component variable
        named `vname`  [defualt 'va'].
    IF_thres : float
        When the ratio of mean wind speed to speed of the mean flow is greater
        than this threshold value, the Incoherent Flow (IF) value of ls_dirn
        will be assigned. Default is 1.5
    WF_thres : float
        On days not classified as IF, if the wind speed is below this
        threshold value the timestep will be assigned the Weak Flow ls_dirn
        value. Default is 2.5 [m/s]

    Returns
    -------
    wind : xarray Dataset
        Contains wind vector properties described

    """
    wind = u
    wind["va"] = v[vname]
    wind["speed"] = np.sqrt(wind.ua ** 2 + wind.va ** 2)
    wind["wdirn"] = calc_wdirn(u[uname], v[vname])

    # average of the wind speed
    wind["ave_speed"] = wind.speed.mean(dim=["latitude", "longitude"])

    # wind speed of the mean flow
    wind["muflow_speed"] = wind.ave_speed
    wind["muflow_speed"].data = calc_mean_wspd(wind.ua, wind.va)

    # direction of the meanflow

    wind["muflow_dirn"] = wind.ave_speed  # dummy dataarray with time dim only
    wind["muflow_dirn"].data = calc_mean_wdirn(wind.ua, wind.va)

    # direction of the large scale flow, where well-defined

    # to ensure contininuity within defined dirn sectors, make N_winds with
    # theta > 337.5, negative

    wind["ls_dirn"] = wind.muflow_dirn.where(
        wind.muflow_dirn < 337.5, other=wind.muflow_dirn - 360
    )

    # artificial wind dirn for incoherent flow days
    wind["ls_dirn"] = wind.ls_dirn.where(
        wind.ave_speed / wind.muflow_speed < IF_thres, other=400
    )
    # artificial wind dirn for weak flow days
    wind["ls_dirn"] = wind.ls_dirn.where(wind.ave_speed > WF_thres, other=350)

    return wind


def prep_rain(file, datecol="DATE", dformat="%Y-%m-%d"):
    """
    Wide-form csv to rainfall xarray Dataset for comparison with wind data.

    Transforms daily rainfall data in a wide-form csv file to an xarray
    Dataset with 6-hourly temporal resolution to compare with 6-hourly wind
    data.

    Parameters
    ----------
    file : str
        The full path to the csv file containing rainfall data
    datecol : str
        The column name in the csv file in which the date is stored. This
        column will be used for the time dimension of the xarray output.
    dateformat : str
        A string specifying the dateformat used in the csv file

    Returns
    -------
    xarray Dataset
        Input transformed to an xarray Dataset with 6-hourly rainfall data.

    """

    rain = pd.read_csv(file)
    rain.rename(columns={datecol: "time"}, inplace=True)
    rain.loc[:, "time"] = pd.to_datetime(rain["time"], format=dformat)

    # rain is daily, wind is 6-hrly; artificially assign rainfall for each 6hrs

    # new dataframes that are separate in memory from original,
    # 1 for each timestep

    rainT0 = rain.copy()
    rainT6 = rain.copy()
    rainT12 = rain.copy()
    rainT18 = rain.copy()

    # uptime time for dataframes

    rainT6.loc[:, "time"] = rainT6["time"] + timedelta(hours=6)
    rainT12.loc[:, "time"] = rainT12["time"] + timedelta(hours=12)
    rainT18.loc[:, "time"] = rainT18["time"] + timedelta(hours=18)

    # correct rainfall to 6hr'ly and 8am SAST-8am SAST recording period

    ncols = len(rainT0.columns)
    nt = len(rain)

    yest = rain.iloc[0: (nt - 1), 1:ncols].values
    today = rain.iloc[1:nt, 1:ncols].values

    # assigning rainfall data to each time step
    rainT0.iloc[1:nt, 1:ncols] = yest / 4
    rainT6.iloc[1:nt, 1:ncols] = yest / 8 + today / 8
    rainT12.iloc[1:nt, 1:ncols] = today / 4
    rainT18.iloc[1:nt, 1:ncols] = today / 4

    rain = pd.concat([rainT0, rainT6, rainT12, rainT18], sort=False)
    rain.sort_values(by="time", inplace=True)  # in chronological order

    rain.iloc[0:2, 1:] = 0  # first 2 timesteps refer to rain recorded
    # for day before first day of record, for which we have no data
    rain.iloc[2, 1:] = yest[0, :] / 4  # 12 and 6pm on day 1
    rain.iloc[3, 1:] = yest[0, :] / 4  # 12 and 6pm on day 1

    rain = rain.melt(id_vars="time", var_name="station", value_name="rain")
    rain.set_index(["time", "station"], inplace=True)  # dimensions for xarray

    return rain.to_xarray()


def meta_index(df, old="NameUsed", new="station", reset=False):
    if reset:
        return df.reset_index().rename(columns={old: new}).set_index(new)
    else:
        return df.rename(columns={old: new}).set_index(new)


def assign_station_meta(data, meta, metavars=["lat", "lon"], sdim="station"):
    """
    Assign meta-data from a pandas DataFrame to an xarray DataArray.

    The selected metadata variables are assigned as co-ordinates to the
    DataArray

    Parameters
    ----------
    data : xarray DataArray
        The array to which metadata should be assigned. Must include point
        data at locations specified by the `sdim` dimension.
    meta : pandas DataFrame
        Data frame with `sdim` as index, containing all the unique values of
        `sdim` found in data and all strings in `metavars` as columns
    metavars : list or array of str, optional
        Variable names for data from `meta` to assign to `data` as
        co-ordinates. The default is ["lat", "lon"].
    sdim : str, optional
        The spatial dimension in data; must also be the index in meta.
        The default is "station".

    Returns
    -------
    data : xarray DataArray
        The original DataArray with requested co-ordinates added

    """
    for v in metavars:
        data = data.assign_coords({v: (sdim, meta.loc[data[sdim], v])})
    return data


# %% assign wind directions to individual named bins

# wind direction bins, including 2 artificials for IF and WF
wbins = 45 * np.arange(11) - 22.5

wnames = ["N", "NE", "E", "SE", "S", "SW", "W", "NW", "WF", "IF"]
# WF = weak flow; IF = (Spatially) Incoherent flow

seaslist = ["DJF", "MAM", "JJA", "SON"]


def assign_wbins(wind):
    """
    Assigns a CWT (wind direction or WF/IF) to a wind Dataset.

    Takese a Dataset output by `prep_wind`. Returns a grouped dataset.

    Parameters
    ----------
    wind : xarray Dataset
        Dataset containing 'ls_dirn' variable. See `prep_wind` for details.

    Returns
    -------
    xarray DatasetGroupBy, grouped over 'ls_dirn_bins'
        The input Dataset, grouped by CWT (labelled as indicated)

    """
    return wind.groupby_bins("ls_dirn", wbins, labels=wnames)


def assign_seas_wbins(wind, timevar="time.season"):
    """
    Seasonal or similar dissaggregated version of `assign_wbins`.

    Returns a diction of grouped Datasets.

    Parameters
    ----------
    wind : xarray Dataset
        Dataset containing 'ls_dirn' variable. See `prep_wind` for details.
    timevar : string, optional
        Time variable to disaggregate grouping by.
        The default is "time.season".

    Returns
    -------
    windseas : dict of xarray DatasetGroupBys, grouped over 'ls_dirn_bins'
        Each element is a seasonal equivalent of the output of `assign_wbins`
        output. Keys are the unique values of `timevar` in `wind`.

    """
    windseas = {}
    seasons = list(set(wind[timevar].values))
    for seas in seasons:
        windseas[seas] = wind.sel(time=wind[timevar] == seas).groupby_bins(
            "ls_dirn", wbins, labels=wnames
        )

    return windseas


def wbins_mean(wind, dim="time"):
    return wind.groupby_bins("ls_dirn", wbins, labels=wnames).mean(dim=dim)


# %% temporal frequency bins


def gr_time_count(
    data, var="ls_dirn", timevar="time.season", countdim=None, grvar="CWT"
):
    """
    Helper function to apply time_count to an xarray binned DatasetGroupBy.

    Defaults to tallying frequencies by season and "ls_dirn_bins", renamed to
    "CWT" on the output,  .

    Parameters
    ----------
    All parameters and defaults are as for time_count, except for the new
    parameter `grvar` which the binned grouping variable is renamed to in the
    output. The default is "CWT".

    Returns
   -------
   xarray DataArray
       A variable with counts along two dimensions: `grvar` and one derived
       from `timevar`.
    """
    varb = var + "_bins"
    return data.apply(time_count, var=var,
                      timevar=timevar, countdim=countdim).rename(
        {varb: grvar}
    )


def time_count(data, var="ls_dirn", timevar="time.season", countdim=None):
    """
    A basis funciton to be applied using `gr_time_count` to tally frequencies.

    Takes in a Dataset grouped by bins of `var`, as ouput, for example, by
    `assign_wbins()`. Returns the number of occurrences of each bin for
     each unique value of `timevar` in `data`.


    Parameters
    ----------
    data : xarray Dataset
        Dataset grouped over bins of `var`
    var : str, optional
        Name of the variable that bins are derived from.
        The default is "ls_dirn".
    timevar : str, optional
        A temporal variable to tally within. The default is "time.season".
    countdim : str or None, optional
        Dimension to count along. If None, tally along all dimensions.
        The default is None.

    Returns
    -------
    See gr_time_count: this function does not produce valid output when
    employed directly.
    """
    if countdim is None:
        return data.groupby(timevar).count(dim=xr.ALL_DIMS)[var]
    else:
        return data.groupby(timevar).count(dim=countdim)[var]


def gr_time2_count(
    data, gr1="syr", gr2="time.season", var="ls_dirn", countdim=None,
    grvar="CWT"
):
    """
    As with gr_time_count, but tallies over an additional grouping variable.

    This function applies `time2_count()`

    Tallies over the binned grouping variable and two more grouping variables.
    Defaults are as for `gr_time_count()`, with the additional grouping
    variable defaulting to "syr". The two grouping variables are named
    `gr1` and `gr2`.
    """
    varb = var + "_bins"
    return data.apply(
        time2_count, gr1=gr1, gr2=gr2, var=var, countdim=countdim, grvar=grvar
    ).rename({varb: grvar})


def time2_count(
    data, gr1="syr", gr2="time.season", var="ls_dirn", countdim=None,
    grvar="CWT"
):
    """
    As with `time_count` but for two grouping variables, to be run with
    `gr_time2_count()`. Where no occurrences are counted, nan is returned.
    """
    data = data.groupby(gr1).apply(time_count, timevar=gr2, var=var,
                                   countdim=countdim)
    return data.where(~(data.isnull()), other=0)  # 0 counts are nan'd


def seas_len(wind, timevar="time.season"):
    """
    Calculate the number of entries in a dataset for each value of `timevar`.

    Parameters
    ----------
    wind : xarray Dataset
        The dataset to tally in.
    timevar : str, optional
        The variable counts are to be disaggregated by.
        The default is "time.season".

    Returns
    -------
    lendict : dict of length 1 xarray DataArrays
        Keys are the unique values of `timevar` in `wind`. The values are
        the corresponding number of timesteps.

    """
    lendict = {}
    seasons = list(set(wind[timevar].values))
    for seas in seasons:
        lendict[seas] = (wind[timevar] == seas).sum()
    return lendict


def gr_mean(wdata, gr, dim=None):
    return wdata.groupby(gr).mean(dim=dim)


# %% wind-dirn rainfall relationship


def dirn_contrib(wdirn, wdirn_freq, rainvar="rain", binvar="ls_dirn",
                 cwtname="CWT"):
    """
    Calculates the proportion of rainfall ocurring under each CWT.

    In addition, returns the mean rainfall occurring under each CWT, and the
    mean under that CWT relative to the overall mean rainfall at that location.

    Parameters
    ----------
    wdirn : xarray Dataset
        Dataset with rainfall in `rainvar`.
    wdirn_freq : xarray DataArray
        Frequencies of each value of the binned grouping variable.
    rainvar : str, optional
        Name of the varible in `wdirn` containing rainfall data to aggregate.
        The default is "rain".
    binvar : str, optional
        The variable used to construct the binned grouping variable.
        The default is "ls_dirn".
    cwtname : str, optional
        New dimension name to replace bin variable. The default is "CWT".

    Returns
    -------
    rain_dirbins : xarray DataArray
        Mean rainfall by CWT.
    norm_rain_dirbins : xarray DataArray
        Normalised (by overall rain rate), rain rate by CWT.
    contrib_rain_dirbins : xarray DataArray
        Proportion of rainfall occurring under each CWT.

    """
    rain_dirbins = (wdirn.mean(dim="time")[rainvar]).rename({binvar + "_bins":
                                                             cwtname})
    contrib_rain_dirbins = rain_dirbins * wdirn_freq
    totals = contrib_rain_dirbins.sum(dim=cwtname)
    norm_rain_dirbins = rain_dirbins / totals
    contrib_rain_dirbins = contrib_rain_dirbins / totals
    return rain_dirbins, norm_rain_dirbins, contrib_rain_dirbins


def dirn_contrib_seas(
    wdirn_dict,
    seasfreq_arr,
    seasdim="season",
    rainvar="rain",
    binvar="ls_dirn",
    cwtname="CWT",
):
    """
    Seasonally disaggreated version of `dirn_contrib()`.

    By default, used to calculate the contribution of each CWT to seasonal
    rainfall totals by season.

    Parameters
    ----------
    wdirn_dict : dict of DataSetGroupBy
        Grouped datasets should be grouped over a binned variable. Datasets
        should also contain a variable from which rainfall contributions.
        Usually the output of `assign_seas_wbins()`
    seasfreq_arr : 2-D xarray DataArray or similar
        Contains the total occurrence of each unique value of
        `cwtname` in each unique instance of `seasdim`. In the default case
        the frequency of each CWT/wind direction bin in each season.
    seasdim : str, optional
        Dimension to disaggregate by. The default is "season".
    rainvar : str, optional
        The name of the variable in each Dataset which contains data from
        which contributions are to be derived. The default is "rain".
    binvar : str, optional
        The variable binned by. The default is "ls_dirn".
    cwtname : str, optional
        New dimension name to apply to the binning aggregation variable.
        The default is "CWT".

    Returns
    -------
    seas_rain : dict of xarray DataArrays
        Dictionary with `seasdim` values as keys and values as in
        `dirn_contrib`
    seas_norm_rain : dict of xarray DataArrays
        As above.
    seas_contrib : dict of xarray DataArrays
        As above.

    """
    seaslist = list(wdirn_dict.keys())
    seas_contrib = {}
    seas_rain = {}
    seas_norm_rain = {}
    for seas in seaslist:
        seas_rain[seas] = (wdirn_dict[seas].mean(dim="time")[rainvar]).rename(
            {binvar + "_bins": cwtname}
        )
        seas_contrib[seas] = seas_rain[seas] * seasfreq_arr.sel({seasdim:
                                                                 seas})
        totals = seas_contrib[seas].sum(dim=cwtname)
        seas_norm_rain[seas] = seas_rain[seas] / totals
        seas_contrib[seas] = seas_contrib[seas] / totals
    return seas_rain, seas_norm_rain, seas_contrib


# %% jacknife versions of above functions


def jk_dirn(
    wind,
    rainvar="rain",
    binvar="ls_dirn",
    cwtname="CWT",
    tvar="time.year",
    yrvar="year"
):
    """
    For each year, return the mean rainfall by CWT in all other years.

    This function is intended to allow "leave-one-out" "prediction" of rainfall
    for a particular year (or similar time-interval) without bias introduced
    by the rainfall-CWT relationship that occurred during that year.

    Parameters
    ----------
    wind : xarray Dataset
        Dataset containing rainfall and CWT variables.
    rainvar : str, optional
        Name of the rainfall variable in `wind`. The default is "rain".
    binvar : str, optional
        Binned grouping variable within which to compute means.
        The default is "ls_dirn".
    cwtname : str, optional
        Dimension name to use in the output for the binned grouping variable.
        The default is "CWT".
    tvar : str, optional
        time variable to perform "leave-one-out" means on.
        The default is "time.year".
    yrvar : str, optional
        dimension name of the year variable in the output.
        The default is "year", corresponding to `tvar='time.year'`.

    Returns
    -------
    xarray DataArray
        Leave-onoe-out mean values of `wind`[`rainvar`] for each year and CWT.

    """

    wind = wind[[rainvar, binvar]]  # only bit we're actually interested in

    # create a dummy dataset of the right shape with dimnames
    yr_rain = wind.groupby(tvar).apply(wbins_mean)[rainvar]
    yrs = np.unique(wind[tvar])

    for yr in yrs:
        # use all other yrs
        yr_rain.loc[{yrvar: yr}] = wbins_mean(wind.sel(time=wind[tvar] != yr)
                                              )[rainvar]

    return yr_rain.rename({binvar + "_bins": cwtname})


def jk_dirn_seas(
    wind, seasdim="season", tvar="syr", rainvar="rain", binvar="ls_dirn",
    cwtname="CWT"
):
    """
    Seasonal disaggregation of `jk_dirn`, by "season year: Dec-Nov"

    Parameters
    ----------
    As in `jk_dirn` except:
    seasdim : str, optional
        Name of disaggregation dimension in `wind`. The default is "season".
    tvar : str, optional
        As for `jk_dirn`, but the default is "syr" for contiguous seasonal
        blocks.

    Returns
    -------
    As with `jk_dirn` with an additional `seasdim`

    """

    # only bit we're actually interested in
    wind = xrc.assign_syr(wind[[rainvar, binvar]])

    # create a dummy dataset of the right shape with dimnames
    yr_rain = wind.groupby_bins(binvar, wbins, labels=wnames).apply(gr_mean,
                                                                    gr=tvar)[
        rainvar
    ]
    yrs = np.unique(wind[tvar])
    seas_rain = {}
    seaslist = np.unique(wind["time." + seasdim])
    for seas in seaslist:
        for yr in yrs:
            seldict = {"time": (wind[tvar] != yr) &
                       (wind["time." + seasdim] == seas)}
            yr_rain.loc[{tvar: yr}] = wbins_mean(wind.sel(seldict))[rainvar]

        seas_rain[seas] = yr_rain

    return dict_2_xr(seas_rain, "season").rename({binvar + "_bins": cwtname})


# %% combined sector and seasonal contributions datasets


def combined_contrib(contrib, wdirns, cwtname="CWT"):
    """
    Adds proportional data along a specified dimension.

    Parameters
    ----------
    contrib : xarray DataArray
        Proportional data along a `cwtname` dimension.
    wdirns :  list or array-like of str
        The wind directions/CWTs over which to sum.
    cwtname :  str, optional
        The name of the dimension along which to aggregate.
        The default is "CWT".

    Returns
    -------
    xarray DataArray
        Aggregation of input `contrib`.

    """
    return contrib.sel({cwtname: wdirns}).sum(dim=cwtname)


def combined_contrib_seas(contrib, wdirns, dimname="season", cwtname="CWT"):
    """
    Seasonally or similarly disaggregated verson of `combined_contrib()`

    Parameters
    ----------
    contrib : dict of xarray DataArrays
        Key values are used for `dimname` in the output. Each entry is as for
        `contrib` in `combined_contrib()`
    wdirns : list or array-like of str
        The wind directions/CWTs over which to sum.
    dimname : str, optional
        Additional dimension to disaggregate by. The default is "season".
    cwtname :  str, optional
        The name of the dimension along which to aggregate.
        The default is "CWT".

    Returns
    -------
    xarray DataArray
        Sums of proportional data

    """
    contrib = dict_2_xr(contrib, dimname)
    return combined_contrib(contrib, wdirns, cwtname=cwtname)


def combined_mean(rain, freq, wdirns, cwtname="CWT"):
    """
    Weighted mean across a selection of CWTs/wind direction bins.

    Takes in a rainfall, rainfall contribution or similar variable, with
    values along a `cwtname` dimension.

    Parameters
    ----------
    rain : xarray DataArray
        Continnuous variable. Must have a `cwtname` dimension.
    freq : xarray DataArray
        Frequency of occurrence of each unique value of `cwtname`
    wdirns : list or array-like of str
        The wind directions/CWTs over which to compute the mean.
    cwtname : str, optional
        The name of the dimension along which to aggregate.
        The default is "CWT".

    Returns
    -------
    xarray DataArray
        The weighted means.

    """
    rain = rain.sel({cwtname: wdirns})
    freq = freq.sel({cwtname: wdirns})
    return xr.dot(rain, freq) / freq.sum(dim=cwtname)


def combined_mean_seas(rain, freq, wdirns, dimname="season", cwtname="CWT"):
    """
    Weighted mean by wind directions, with season or similar dimension

    Parameters
    ----------
    See combined mean.
    `rain` and `freq` can be dictionaries of DataArrays, where the dictionary
    keys become the values of `dimname` in the output, or DataArrays
    with with `dimname` as a dimension. All arrays must have `cwtname` as a
    dimension
    One addional parameter:
    dimname : str, optional
        The dimension along which to disaggregate means.
        The default is "season".

    Returns
    -------
    seas_mean : xarray DataArray
        Means along the `cwtname` and `dimname` dimensions.
    """


    if rain is dict:
        rain = dict_2_xr(rain, dimname)
    if freq is dict:
        freq = dict_2_xr(freq, dimname)
    seasons = list(set(rain[dimname].values))
    seas_mean = xr.zeros_like(rain.mean(dim=cwtname))  # initialise with
    # CWT dimension removed

    # select out the wind dirns/CWTs we're interested in:
    rain = rain.sel({cwtname: wdirns})
    freq = freq.sel({cwtname: wdirns})
    for seas in seasons:
        rain_seas = rain.sel({dimname: seas})
        freq_seas = freq.sel({dimname: seas})
        seas_mean.loc[{dimname: seas}] = xr.dot(rain_seas,
                                                freq_seas) / freq_seas.sum(
            dim=cwtname
        )
    return seas_mean


#     return np.dot(rain, freq)/freq.sum(dim='wdirns_ls_bins')


# %% conversion


def dict_2_xr(xrdict, concat_dim):
    """
    Converts a dictionary of xarray objects to a single xarray object.

    The objects are concatenated along a new dimension, whose name is
    specified by `concatdim` and whose values are the keys of `xrdict`

    Parameters
    ----------
    xrdict : dict of xarray objects (Datasets or DataArrays)
        A dictionary containing the xarray objects to be concetaned.
    concat_dim : str
        The name of the new dimension along which the arrays are to be
        concatenated.

    Returns
    -------
    da : same as elements of xrdict
        The concatenated xarray object.

    """
    da = xr.concat([xrdict[i] for i in xrdict.keys()], dim=concat_dim)
    da[concat_dim] = list(xrdict.keys())
    return da


# %% direction of maximum rainfall/ contribution


def max_wdirn(data, cwtname="CWT"):
    """
    Returns the wind direction (CWT) for which the variable attains its max.

    This function is intended to be used with a rainfall, rainfall intensity,
    rainfall contribution, or similar variable, with `cwtname` as a dimension.

    Parameters
    ----------
    data : xarray DataArray
        Continuous variable with `cwtname` dimension and possibly others.
    cwtname : str, optional
        Dimension to aggregate over. The default is "CWT".

    Returns
    -------
    max_dirn : xarray DataArray
        Array of strings of the name of the CWT (or similar) for which the
        maximum value is attained.

    """
    # add nonsense to ensure it's not seen as co-ord
    max_dirn = data.station + "1"
    stationslist = data.station.values
    wbins = data[cwtname].values
    for st in stationslist:
        max_dirn.loc[st] = wbins[data.sel(station=st).argmax()]
    return max_dirn
