#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Sep 22 16:54:46 2019
Run from <root>/py_scripts/
@author: stefaan
"""

# %% Imports

import pandas as pd
import xarray as xr
import numpy as np
import fn.wind_dirn as wd
import fn.wind_plots as wp
import fn.basemap_plots as bmp
import fn.station_clusters as sc
import fn.xrclim_fns as xrc

# %% Wind data import and preparation

# paths to netcdf wind data

path = "../data/wind/"  # directory in which wind data are kept

ufile = "ERA5_u900.nc"  # location of u wind data
vfile = "ERA5_v900.nc"  # location of v wind data

# nc file input

u900 = xr.open_dataset(path + ufile).isel(level=0)
v900 = xr.open_dataset(path + vfile).isel(level=0)

# new x-array dataset

wind900 = wd.prep_wind(u900, v900)
wmeta = "ERA5"  # meta data for wind data

# constants
rd_thres = 1  # mm/day
# mean days per met season
seaslen_mu = {"DJF": 90.25, "MAM": 92, "JJA": 92, "SON": 91}

# %% pandas rainfall dataframe

# The rainfall dataframe should be read in from a wide-format daily rainfall
# csv file, with one column for the date, and then one column per station
# in which daily rainfall in mm is recorded. This script requires gap-free
# data, so some form of infilling will need to be performed to obtain these
# data. The "rainfall.csv" included below uses only freely available data
# from the DWS and has been infilled but not cleaned, is not very dense
# and consequently is not a very high quality dataset, but it can be used to
# demonstrate all functionality of this script.
file = "../data/rainfall.csv"

xrain = wd.prep_rain(file, datecol="date")  # very RAM intensive

# %% Combine wind and rain data

# for rain, select out the times for which wind data are available
xrain = xrain.sel(time=slice(wind900.time[0], wind900.time[-1]))

# similar for wind
wind900 = wind900.sel(time=slice(xrain.time[0], xrain.time[-1]))

wind900 = wind900.merge(xrain)  # combine rainfall and wind datasets

# 1mm/day threshold <=> 0.25mm/6hr
wind900["rainday"] = wind900.rain >= rd_thres / 4

# for seasonal analysis
wdirn900syr = wd.assign_wbins(xrc.assign_syr(wind900))
wdirn900 = wd.assign_wbins(wind900)

# %% get metadata

metafile = "../metadata/WRZ_meta.csv"
meta = pd.read_csv(metafile)

# set station as index for safe comparison and name it station
meta = wd.meta_index(meta)

slist = meta.index.values  # station names

# %% Frequency Bar plotting

# frequency plot of dirn groups: rename them as 'CWT'
wdirn_count = wdirn900.count(dim=xr.ALL_DIMS).ls_dirn.rename({"ls_dirn_bins":
                                                              "CWT"})

freq_wdirn900 = wdirn_count / len(xrain.time)

nyrs = len(xrain.time) / (4 * 365.25)  # years of data available

wp.wdir_freq_plot(wdirn900, figmeta=wmeta + "_annual", period="year", nyr=nyrs)

# %% frequency of occurrence

freq_wdirn900ann = wd.gr_time_count(wdirn900, timevar="time.year")
freq_wdirn900seas = wd.gr_time_count(wdirn900)
freq_wdirn900ann_seas = wd.gr_time2_count(wdirn900syr)

# %% wind direction contribution to rainfall by season and year

rain_dirn, normrain_dirn, contrib_rain_dirn = wd.dirn_contrib(wdirn900,
                                                              freq_wdirn900)

rd_dirn, normrd_dirn, contrib_rd_dirn = wd.dirn_contrib(
    wdirn900, freq_wdirn900, rainvar="rainday"
)

# seasonally

seas_wdirn = wd.assign_seas_wbins(wind900)

seas_lens = wd.seas_len(wind900)

(rain_seas, normrain_seas, contrib_seas) = wd.dirn_contrib_seas(
    seas_wdirn, freq_wdirn900seas
)

rd_seas, normrd_seas, contrib_rd_seas = wd.dirn_contrib_seas(
    seas_wdirn, freq_wdirn900seas, rainvar="rainday"
)
# %% annually

year_lens = wd.seas_len(wind900, timevar="time.year")

wdirn_yr = wd.assign_seas_wbins(wind900, timevar="time.year")

rain_ann, norm_rain_ann, contrib_ann = wd.dirn_contrib_seas(
    wdirn_yr, freq_wdirn900ann, seasdim="year"
)

rd_ann, norm_rd_ann, contrib_rd_ann = wd.dirn_contrib_seas(
    wdirn_yr, freq_wdirn900ann, seasdim="year", rainvar="rainday"
)

rain_ann = wd.dict_2_xr(rain_ann, "year")
norm_rain_ann = wd.dict_2_xr(norm_rain_ann, "year")
rd_ann = wd.dict_2_xr(rd_ann, "year")
norm_rd_ann = wd.dict_2_xr(norm_rd_ann, "year")

contrib_ann = wd.dict_2_xr(contrib_ann, "year")  # convert to xarray
contrib_rd_ann = wd.dict_2_xr(contrib_rd_ann, "year")

# %% jacknife rain assessments for `prediction'

jk_ann = wd.jk_dirn(wind900)
jk_ann = jk_ann.sel(year=jk_ann.year < 2019)  # exclude incomplete yr

jk_seas = wd.jk_dirn_seas(wind900)

# %% seasonal frequency plots

seasonslist = list(set(xrain["time.season"].values))

# seasyrs = [39+59/90, 40, 40, 39+61/91] # years of data per season
seasyrs = {
    seas: (0.25 * seas_lens[seas] / seaslen_mu[seas]).values
    for seas in seaslen_mu.keys()
}

# 0.25 for the 4 time periods per day

wp.seas_wdir_freq_plot(seas_wdirn, "seasonal", figmeta=wmeta + "_",
                       nyr=seasyrs)

# %% sector contributions

contrib_W_NW = wd.combined_contrib(contrib_rain_dirn, ["W", "NW"])
contrib_S_SE = wd.combined_contrib(contrib_rain_dirn, ["S", "SE"])
contrib_W_all = wd.combined_contrib(contrib_rain_dirn, ["SW", "W", "NW"])
contrib_IWF = wd.combined_contrib(contrib_rain_dirn, ["IF", "WF"])
contrib_E_NE = wd.combined_contrib(contrib_rain_dirn, ["E", "NE"])
contrib_S_all = wd.combined_contrib(contrib_rain_dirn, ["SW", "S", "SE"])
contrib_E_all = wd.combined_contrib(contrib_rain_dirn, ["E", "NE", "SE"])
contrib_N_all = wd.combined_contrib(contrib_rain_dirn, ["N", "NE", "NW"])

# seasonally

contrib_W_NW_seas = wd.combined_contrib_seas(contrib_seas, ["W", "NW"])
contrib_S_SE_seas = wd.combined_contrib_seas(contrib_seas, ["S", "SE"])
contrib_W_all_seas = wd.combined_contrib_seas(contrib_seas, ["SW", "W", "NW"])
contrib_IWF_seas = wd.combined_contrib_seas(contrib_seas, ["IF", "WF"])
contrib_E_NE_seas = wd.combined_contrib_seas(contrib_seas, ["E", "NE"])
contrib_S_all_seas = wd.combined_contrib_seas(contrib_seas, ["SW", "S", "SE"])
contrib_E_all_seas = wd.combined_contrib_seas(contrib_seas, ["E", "NE", "SE"])
contrib_N_all_seas = wd.combined_contrib_seas(contrib_seas, ["N", "NE", "NW"])

contrib_seas = wd.dict_2_xr(contrib_seas, "season")  # convert to xarray
normrain_seas = wd.dict_2_xr(normrain_seas, "season")
rain_seas = wd.dict_2_xr(rain_seas, "season")

contrib_rd_seas = wd.dict_2_xr(contrib_rd_seas, "season")
normrd_seas = wd.dict_2_xr(normrd_seas, "season")
rd_seas = wd.dict_2_xr(rd_seas, "season")

# %% sector normalised rainfall

W_NW_norm = wd.combined_mean(normrain_dirn, freq_wdirn900, ["W", "NW"])
S_SE_norm = wd.combined_mean(normrain_dirn, freq_wdirn900, ["S", "SE"])
W_all_norm = wd.combined_mean(normrain_dirn, freq_wdirn900, ["SW", "W", "NW"])
IWF_norm = wd.combined_mean(normrain_dirn, freq_wdirn900, ["IF", "WF"])
E_NE_norm = wd.combined_mean(normrain_dirn, freq_wdirn900, ["E", "NE"])
S_all_norm = wd.combined_mean(normrain_dirn, freq_wdirn900, ["SW", "S", "SE"])
E_all_norm = wd.combined_mean(normrain_dirn, freq_wdirn900, ["E", "NE", "SE"])
N_all_norm = wd.combined_mean(normrain_dirn, freq_wdirn900, ["N", "NE", "NW"])

# seasonally

W_NW_norm_seas = wd.combined_mean_seas(normrain_seas, freq_wdirn900seas,
                                       ["W", "NW"])
S_SE_norm_seas = wd.combined_mean_seas(normrain_seas, freq_wdirn900seas,
                                       ["S", "SE"])
W_all_norm_seas = wd.combined_mean_seas(
    normrain_seas, freq_wdirn900seas, ["SW", "W", "NW"]
)
IWF_norm_seas = wd.combined_mean_seas(normrain_seas, freq_wdirn900seas,
                                      ["IF", "WF"])
E_NE_norm_seas = wd.combined_mean_seas(normrain_seas, freq_wdirn900seas,
                                       ["E", "NE"])
S_all_norm_seas = wd.combined_mean_seas(
    normrain_seas, freq_wdirn900seas, ["SW", "S", "SE"]
)
E_all_norm_seas = wd.combined_mean_seas(
    normrain_seas, freq_wdirn900seas, ["E", "NE", "SE"]
)
N_all_norm_seas = wd.combined_mean_seas(
    normrain_seas, freq_wdirn900seas, ["N", "NE", "NW"]
)

# %% wind directions with highest rainfall

max_dirn = wd.max_wdirn(normrain_dirn)
# direction with highest contribution
max_contrib_dirn = wd.max_wdirn(contrib_rain_dirn)

max_dirn_season = xr.DataArray()
for i, seas in enumerate(seasonslist):
    temp = wd.max_wdirn(normrain_seas.sel(season=seas))
    temp["season"] = seas
    if i == 0:
        max_dirn_seas = temp
    else:
        max_dirn_seas = xr.concat([max_dirn_seas, temp], dim="season")
    temp = wd.max_wdirn(contrib_seas.sel(season=seas))
    if i == 0:
        max_contrib_dirn_seas = temp
    else:
        max_contrib_dirn_seas = xr.concat([max_contrib_dirn_seas, temp],
                                          dim="season")

# %% spatial plotting

# spatial bounds

latN = -31
latS = -35.4
lonW = 17
lonE = 21.7

bounds = [lonW, lonE, latS, latN]

# %% contribution by CWT combinations

bmap, fig = bmp.station_basemap(bounds=bounds, width=270, height=250)

# contribution plots

wp.annual_contrib_plot(
    data=contrib_W_NW,
    vlims=[0.25, 0.8],
    figmeta=wmeta + "W_NW",
    meta=meta,
    bmap=bmap,
    desc="W or NW flow",
    fig=fig,
)

wp.annual_contrib_plot(
    data=contrib_W_all,
    vlims=[0.4, 0.9],
    meta=meta,
    figmeta=wmeta + "W_all",
    bmap=bmap,
    desc="SW, W or NW flow",
    fig=fig,
)

wp.annual_contrib_plot(
    data=contrib_S_SE,
    meta=meta,
    vlims=[0, 0.4],
    figmeta=wmeta + "S_SE",
    desc="S or SE flow",
    bmap=bmap,
    fig=fig,
)

wp.annual_contrib_plot(
    data=contrib_E_NE,
    meta=meta,
    vlims=[0, 0.2],
    figmeta=wmeta + "E_NE",
    desc="E or NE flow",
    bmap=bmap,
    fig=fig,
)

# %% CWT with max rain

bmap, fig = bmp.station_basemap(bounds=bounds, width=250, height=250)

# assign lon and lat for discrete_var_map
max_dirn = wd.assign_station_meta(max_dirn, meta)

max_contrib_dirn = wd.assign_station_meta(max_contrib_dirn, meta)

max_dirn_seas = wd.assign_station_meta(max_dirn_seas, meta)

max_contrib_dirn_seas = wd.assign_station_meta(max_contrib_dirn_seas, meta)

bmp.discrete_var_map(max_dirn, clabel="CWT", descr=wmeta + "dirn_maxrain",
                     bmap=bmap)

for seas in seasonslist:
    bmp.discrete_var_map(
        max_dirn_seas.sel(season=seas),
        clabel="CWT",
        descr=wmeta + "dirn_maxrain_" + seas,
        bmap=bmap,
    )

# direction with highest rainfall contrib

bmp.discrete_var_map(
    max_contrib_dirn, bmap=bmap, descr=wmeta + "dirn_maxcontrib", clabel="CWT"
)

for seas in seasonslist:
    bmp.discrete_var_map(
        max_contrib_dirn_seas.sel(season=seas),
        bmap=bmap,
        descr=wmeta + "dirn_maxcontrib_" + seas,
        clabel="CWT",
    )

# %% by season

bmap, fig, axdict = bmp.station_basemap_fg(
    bounds=bounds,
    width=170,
    faceter=seasonslist,
    height=300,
    titles=seasonslist,
    nlons=3,
)

wp.seasonal_contrib_plot(
    data=contrib_W_NW_seas,
    meta=meta,
    fig=fig,
    bmap=bmap,
    axesdict=axdict,
    vlims=[0.1, 0.9],
    figmeta=wmeta + "W_NW",
    desc="W or NW flow",
)

wp.seasonal_contrib_plot(
    data=contrib_S_SE_seas,
    meta=meta,
    fig=fig,
    bmap=bmap,
    axesdict=axdict,
    vlims=[0.1, 0.55],
    figmeta=wmeta + "S_SE",
    desc="S or SE flow",
)

wp.seasonal_contrib_plot(
    data=contrib_E_NE_seas,
    meta=meta,
    fig=fig,
    bmap=bmap,
    axesdict=axdict,
    vlims=[0, 0.3],
    figmeta=wmeta + "E_NE",
    desc="E or NE flow",
)

# %% by wind direction

wbins = freq_wdirn900ann.CWT.values
bmap, fig, axdict = bmp.station_basemap_fg(
    bounds=bounds,
    width=365,
    faceter=wbins,
    height=200,
    titles=wbins,
    ncol=5,
    nrow=2,
    nlats=6,
    nlons=3,
)

wp.wdirn_rain_plot(
    normrain_dirn,
    meta,
    unit="relative to mean",
    bmap=bmap,
    fig=fig,
    axesdict=axdict,
    vlims=[0.33, 3],
    figmeta=wmeta + "norm_rain",
    log=True,
    cbarlab="mean rainfall under CWT",
)

wp.wdirn_rain_plot(
    rd_dirn,
    meta,
    unit="1",
    bmap=bmap,
    fig=fig,
    axesdict=axdict,
    vlims=[0.01, 0.7],
    figmeta=wmeta + "rainday",
    log=True,
    cbarlab="proportion of days with rain by CWT",
)

wp.wdirn_rain_plot(
    normrd_dirn,
    meta,
    unit="relative to mean",
    bmap=bmap,
    fig=fig,
    axesdict=axdict,
    vlims=[0.25, 4],
    figmeta=wmeta + "norm_rainday",
    log=True,
    cbarlab="relative rainday frequency",
)

wp.wdirn_rain_plot(
    100 * contrib_rain_dirn,
    meta,
    unit="%",
    bmap=bmap,
    fig=fig,
    axesdict=axdict,
    figmeta=wmeta + "contrib",
    vlims=[0, 40],
    cbarlab="rainfall under CWT",
)

for seas in seasonslist:
    wp.wdirn_rain_plot(
        normrain_seas.sel(season=seas),
        meta,
        bmap=bmap,
        fig=fig,
        unit="relative to mean",
        axesdict=axdict,
        figmeta=wmeta + "norm_rain_" + seas,
        vlims=[0.25, 4],
        cbarlab="mean rainfall by CWT",
        log=True,
    )

    wp.wdirn_rain_plot(
        100 * contrib_seas.sel(season=seas),
        meta,
        fig=fig,
        unit="%",
        bmap=bmap,
        vlims=[0, 40],
        axesdict=axdict,
        figmeta=wmeta + "contrib_" + seas,
        cbarlab="contribution to rainfall by CWT",
    )

# %% trend plots: evolution of frequency of CWT occurrence
# rainfall data are not required for this code cell

freq_W_NW_byyr = wp.freq_trend_plot(freq_wdirn900ann,
                                    wmeta + "_W_NW_annual_freq")
freq_W_all_byyr = wp.freq_trend_plot(
    freq_wdirn900ann, dirns=["SW", "W", "NW"],
    descr=wmeta + "_W_all_annual_freq"
)
freq_S_SE_byyr = wp.freq_trend_plot(
    freq_wdirn900ann, dirns=["S", "SE"], descr=wmeta + "_S_SE_annual_freq"
)
freq_E_NE_byyr = wp.freq_trend_plot(
    freq_wdirn900ann, dirns=["E", "NE"], descr=wmeta + "_E_NE_annual_freq"
)
freq_IWF_byyr = wp.freq_trend_plot(
    freq_wdirn900ann, dirns=["IF", "WF"], descr=wmeta + "_IF_WF_annual_freq"
)

wp.freq_trend_plot_fg(
    freq_wdirn900ann, descr=wmeta + "_wdirn_ann_freq", facetvar="CWT", ncol=5
)

freq_W_NW_byseas = {}
freq_W_all_byseas = {}
freq_S_SE_byseas = {}
freq_E_NE_byseas = {}
freq_IWF_byseas = {}

for seas in seasonslist:
    freq_seas = freq_wdirn900ann_seas.sel(season=seas)
    freq_W_NW_byseas[seas] = wp.freq_trend_plot(
        freq_seas, wmeta + "_W_NW_freq_" + seas, period="season"
    )

    freq_W_all_byseas[seas] = wp.freq_trend_plot(
        freq_seas,
        dirns=["SW", "W", "NW"],
        descr=(wmeta + "_W_all_freq_" + seas),
        period="season",
    )

    freq_S_SE_byseas[seas] = wp.freq_trend_plot(
        freq_seas,
        dirns=["S", "SE"],
        period="season",
        descr=(wmeta + "_S_SE_freq_" + seas),
    )

    freq_E_NE_byseas[seas] = wp.freq_trend_plot(
        freq_seas,
        dirns=["E", "NE"],
        period="season",
        descr=(wmeta + "_E_NE_freq_" + seas),
    )

    freq_IWF_byseas[seas] = wp.freq_trend_plot(
        freq_seas,
        dirns=["IF", "WF"],
        period="season",
        descr=(wmeta + "_IF_WF_freq_" + seas),
    )

    wp.freq_trend_plot_fg(
        freq_seas,
        descr=wmeta + "_freq_wdirn900" + seas,
        facetvar="CWT",
        ncol=5,
        xrename={"syr": "year"},
    )

freq_W_NW_byseas = wd.dict_2_xr(freq_W_NW_byseas, "season")
freq_W_all_byseas = wd.dict_2_xr(freq_W_all_byseas, "season")
freq_S_SE_byseas = wd.dict_2_xr(freq_S_SE_byseas, "season")
freq_E_NE_byseas = wd.dict_2_xr(freq_E_NE_byseas, "season")
freq_IWF_byseas = wd.dict_2_xr(freq_IWF_byseas, "season")

# wdirn_seas_freq = wd.dict_2_xr(wdirn_seas_freq, 'season')


# %% contribution clusters

# The data required for the following clustering step, if stored in
# ../data/wind/contrib_rain_dirn.nc can be read in and used without running
# the prior steps; run the "imports" cell in this script, uncomment the line
# below and then run this cell.

# contrib_rain_dirn = xr.open_dataset(path + "contrib_rain_dirn.nc")

sc.wdirn_cluster(
    contrib_rain_dirn, meta, descr=wmeta + "_wind900_", clustmin=3, clustmax=5,
    opt=3
)
