#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Aug 10 19:42:13 2020
Functions for Basemap, bar and time series plots for CWT analysis.
When written to file, figures are placed in a 'figures' subdirectory of the
working directorym unless the 'figdir' constant is set to a different
directory.
@author: stefaan
"""
# %% imports

import pandas as pd
import xarray as xr
import numpy as np
import matplotlib.pyplot as plt
import fn.wind_dirn as wd
import fn.basemap_plots as bmp
import seaborn as sns

# %%

sns.set(style="darkgrid")
sns.set_context("talk")
figdir = "../figures/"

# %% Qualitative colour palettes

# Hex lists of colours

# Tol colours:

# blue red green yellow cyan purple grey black
Tol_bright = ['#4477AA', '#EE6677', '#228833', '#CCBB44', '#66CCEE',
              '#AA3377', '#BBBBBB', '#000000']

# my alternatives, with a different grey, at the end  (no. 4) and with
# black first

# blue red green yellow cyan purple black grey
Tol_b2 = ['#4477AA', '#EE6677', '#228833', '#CCBB44', '#66CCEE',
          '#AA3377', '#000000', '#555555']

# blue red green grey yellow cyan purple black
Tol_b3 = ['#4477AA', '#EE6677', '#228833', '#555555', '#CCBB44', '#66CCEE',
          '#AA3377', '#000000']

Tol_b1 = ['#000000', '#EE6677', '#CCBB44', '#66CCEE', '#228833', '#555555',
          '#4477AA', '#AA3377']  # black red yellow cyan green grey blue purple

# high contrast

Tol_hc = ['#004488', '#DDAA33', '#BB5566', '#000000']  # blue yellow red black

# orange blue cyan magenta red teal grey black
Tol_vibrant = ['#EE7733', '#0077BB', '#33BBEE', '#EE3377', '#CC3311',
               '#009988', '#BBBBBB', '#000000']

# rose indigo sand green cyan wine teal olive purple pale_grey black
Tol_muted = ['#CC6677', '#332288', '#DDCC77', '#117733', '#88CCEE',
             '#882255', '#44AA99', '#999933', '#AA4499', '#DDDDDD',
             '#000000']
# orange red rose purple wine indigo blue cyan teal green olive sand black
# pale_grey
my15 = ['#EE7733', '#CC3311', '#CC6677', '#AA4499', '#882255', '#332288',
        '#0077BB', '#88CCEE', '#44AA99', '#117733', '#999933', '#DDCC77',
        '#000000', '#DDDDDD']

# my adaptation of Tol_muted for when first colour is to be reserved for the
# ocean, 2nd colour for single entity and other colours come in 2s; last
# colour for NA

# cyan black indigo wine sand olive green teal pale_grey
Tol_seafirst_paired = ['#88CCEE', '#000000', '#332288', '#882255', '#DDCC77',
                       '#999933', '#117733', '#44AA99', '#DDDDDD']

# light_blue orange light_yellow pink light_cyan mint pear olive pale_grey
# black
Tol_light = ['#77AADD', '#EE8866', '#EEDD88', '#FFAABB', '#99DDFF',
             '#44BB99', '#BBCC33', '#AAAA00', '#DDDDDD', '#000000']

# Wong 2011 (nature methods) colours:

# Black, Orange, Sky_Blue, Bluish_Green, Yellow, Blue, Vermillion,
# Reddish purple
Wong = ["#000000", "#E69D00", "#56B4E9", "#009E74", "#F0E442", "#0072B2",
        "#D55C00", "#CC79A7"]

# IBM

IBM = ["#648FFF", "#785EF0", "#DC267F", "FE6100", "FFB000"]

# from davidmathlogic.com/colorblind

col8 = ['tab:orange', 'tab:blue', 'indigo', 'tab:red', 'yellow',
        'tab:grey', 'black', 'tab:cyan']


# %% Bar Plotting functions

def wdir_freq_plot(grdata, figmeta, var='ls_dirn', nyr=40, period="year",
                   xlab="CWT", width=200, height=150):
    """
    Bars of climatological frequency for Dataset grouped by binning variable.

    It is assumed that six-hourly data (4x daily) are supplied.

    Parameters
    ----------
    grdata : xarray DatasetGroupBy
        A dataset grouped by CWT/wind direction bins or similar categorical
        variable to aggregate over.
    figmeta : str
        Description of the data used in the figure to use for naming the saved
        output.
    var : str, optional
        A variable in the dataset to count along. The default is 'ls_dirn'.
    nyr : float, optional
        Number of years of data used. The default is 40.
    period : str, optional
        Duration of . The default is "year".
    xlab : str, optional
        Label of the x-axis. The default is "CWT".
    width : float, optional
        width of figure in mm. The default is 200.
    height : float, optional
        height of figure in mm. The default is 150.

    Returns
    -------
    None.

    """
    # set xlab=None for no xlab
    win = width / 25.4    # width in inches
    hin = height / 25.4   # height in inches

    ax = (grdata.count(dim=xr.ALL_DIMS)[var].to_series() / (4 * nyr)
          ).plot.bar(figsize=(win, hin))  # /4 for 6hr'ly => daily
    ax.set(ylabel="frequency (days/" + period + ")", xlabel=xlab)
    fig = ax.get_figure()
    fig.savefig(figdir + figmeta + 'wdirn' + '_freq.png',
                bbox_inches='tight')
    plt.close('all')


def seas_wdir_freq_plot(grdict, name, figmeta, nyr=40, concatdim='season',
                        var='ls_dirn', xlab="CWT", width=240, height=150,
                        colours=col8):
    """
    Seasonal equivalent of `wdirn_freq_plot().`

    Takes in a dictionary of grouped Datasets. A facetgrid of plots is created.
    The panel titles are derived from the dictionary keys.

    Parameters
    ----------
    As for `wdir_freq_plot()`, except that `grdict` is used in stead of
    `grdata`, `concatdim` and `colours` are added and
    different defaults apply as indicated below:
    grdict : dict of xarray DatasetGroupBy
        Each entry in the dictionary is of the same structure as `grdata`
        parameter for `wdir_freq_plot()`.
    concatdim : str, optional
        The name of variable whose values are the dictionary keys.
        The default is 'season'. Used for the "ylabel".
    xlab : str, optional
        Label on the x-axis. Set xlab to empty string ("") for no xlab;
        None fails as xlab is used as a dimension. The default is "CWT".
    width : float, optional
        The default is 240 (mm).
    height : float, optional
        The default is 150 (mm).
    colours : str or list or array of str
        The default is col8.

    Returns
    -------
    None.

    """
    win = width / 25.4     # width in inches
    hin = height / 25.4    # height in inches

    # seasons with fixed specified number of yrs of data
    if not type(nyr) is dict:
        nyr = {seas: nyr for seas in grdict.keys()}

    # get dictionary of grouped xarray DAs as DAs and then concat into 1 DA.
    xrda = wd.dict_2_xr({seas: (grdict[seas].count(dim=xr.ALL_DIMS)[var] /
                                (4 * nyr[seas]))
                         for seas in grdict.keys()},
                        concatdim).rename({var+'_bins': xlab})

    ax = (xrda.T.to_pandas()).plot.bar(color=colours, figsize=(win, hin))
    # transpose to seasons or equivalent variable by colour
    ax.set(ylabel="frequency (days/" + concatdim + ")")
    fig = ax.get_figure()
    fig.savefig(figdir + figmeta + 'wdirn_' + name + '_freq.png',
                bbox_inches='tight')
    plt.close('all')


def contrib_bars(data, facetvar, figmeta, ncol=2, xlab="CWT",
                 colours=Tol_b1, width=200, height=200):
    """
    Facetgrid-type barplot of the proportion of rainfall occurring by CWT.

    Each facet represents the rainfall contributions of a different value of
    a faceting dimension supplied by the user.

    Parameters
    ----------
    data : 2-D xarray DataArray
        Climatological contribution to annual total for a continuous variable
        (rainfall or similar) by CWT or similar categorical dimension and
        an additional facetting dimension (`facetvar`).
    facetvar : str
        Dimension along which to disaggregate for faceting.
    figmeta : str
        Description of the data used in the figure to use for naming the saved
        output.
    ncol : int, optional
    Number of columns in the facetgrid. The default is 2.
    xlab : str, optional
        Label of the x-axis. The default is "CWT".
    colours : list, or str optional
        list or list-like of colour names or string defining a built-in
        seaborn colourmap. The default is Tol_b1 [defined in this script].
    width : float, optional
        Width of the full plot in mm. The default is 200.
    height : float, optional
        Heigh of the full plot in mm. The default is 200.

    Returns
    -------
    None.

    """
    facetlevs = pd.unique(data[facetvar].values)  # all facets
    nf = len(facetlevs)  # number of facets

    # force to integer to avoid depr. warning
    nrow = np.int(np.ceil(nf / ncol))
    maxval = data.max() * 100  # to %

    colours = sns.color_palette(colours, n_colors=nf)

    fig = plt.figure(figsize=(width / 25, height / 25))

    for i, f in enumerate(facetlevs):
        facet_data = data.sel({facetvar: f}) * 100  # to %
        ax = fig.add_subplot(nrow, ncol, i + 1)
        facet_data.to_series().plot.bar(ax=ax, color=colours[i])
        if (i % ncol == 0):  # left
            ax.set(ylabel="% contribution")
        else:
            ax.set(yticklabels=[])

        if (i >= (nrow - 1) * (nf / nrow)):  # bottom row
            ax.set(xlabel=xlab)
        else:
            ax.set(xlabel=None, xticklabels=[])
        ax.set(title=f)
        ax.set(ybound=(0, maxval+2))

    fig.subplots_adjust(left=0, right=1.5, top=3, bottom=0)
    fig.savefig(figdir+figmeta+'.png', bbox_inches='tight')
    plt.close('all')

# %% Map plotting functions


def seasonal_contrib_plot(data, meta, bmap, fig, axesdict, facetvar='season',
                          vlims=[0, 1], figmeta="W_NW_I", desc="W or NW flow"):
    """
    Maps the observed proportion of rainfall for a CWT at stations by season.

    A short-cut for the special case of the `cts_var_map_fg` function from
    basemap_plots.py where the data to be mapped are the seasonal proportions
    of rainfall occuring under a given CWT or combination of CWTs, by season.

    The plot uses data converted from fractions to percentages.

    Parameters are equivalent to `cts_var_map_fg` except for:
    ----------
    meta : pandas DataFrame
        a list of stations, with "station" index, and 'lat' and 'lon' columns
        containing their latitude and longitude in degrees N and E,
        respectively.
    facetvar : The default is 'season'.
    vlims : The default is [0, 1].
    desc : The default is "W or NW flow".
    figmeta: The default is "W_NW_I".

    Returns
    -------
    None.

    """
    # assign lon and lat for cts_var_map
    data = wd.assign_station_meta(data, meta)

    # 100* to get data in % form
    bmp.cts_var_map_fg(100 * data, facetvar=facetvar, axdict=axesdict,
                       bmap=bmap, descr=figmeta + "_wdirn_contrib" + facetvar,
                       clabel="% rainfall with "+desc, fig=fig,
                       vlims=[v * 100 for v in vlims])


def annual_contrib_plot(data, meta, bmap, fig, vlims=[0, 1],
                        figmeta="W_NW_I", desc="W or NW flow"):
    """
    Maps the observed annual contribution to rainfall occurring at stations.

    A short-cut for the special case of the `cts_var_map` function from
    basemap_plots.py where the data to be mapped are contributions to total
    annual rainfall occurring under a given CWT.

    The plot produced converts fractions (proportions) to percentages.

    Parameters are equivalent to `cts_var_map` except for:

    Parameters
    ----------
    meta : pandas DataFrame
        a list of stations, with "station" index, and 'lat' and 'lon' columns
        containing their latitude and longitude in degrees N and E,
        respectively.
    vlims : The default is [0, 1].
    figmeta: The default is "W_NW_I".
    log : The default is True.

    The following `cts_var_map` are set:
        ``clabel="% rainfall with "``
        ``descr=figmeta + "_wdirn_contrib"``
    Returns
    -------
    None.

    """
    #    bmap : mpl_toolkits.basemap.Basemap
    #        A basemap object, onto which to plot point data.

    # assign lon and lat for cts_var_map
    data = wd.assign_station_meta(data, meta)

    # 100* to get data in % form
    bmp.cts_var_map(100 * data, descr=figmeta + "_wdirn_contrib",
                    bmap=bmap, fig=fig, clabel="% rainfall with " + desc,
                    vlims=[v * 100 for v in vlims])


def rain_plot(data, meta, bmap, fig, vlims=[0.33, 3], log=True,
              unit='relative to mean', desc="W or NW flow", figmeta="W_NW_I"):
    """
    Maps the observed relative rain rate occurring at stations.

    A short-cut for the special case of the `cts_var_map` function from
    basemap_plots.py where the data to be mapped are the relative rain rate.

    Parameters are equivalent to `cts_var_map` except for:

    Parameters
    ----------
    meta : pandas DataFrame
        a list of stations, with "station" index, and 'lat' and 'lon' columns
        containing their latitude and longitude in degrees N and E,
        respectively.
    vlims : The default is [0.33, 3].
    unit : The default is 'relative to mean'.
    desc : The default is "W or NW flow".
    figmeta: The default is "W_NW_I".
    log : The default is True.

    Returns
    -------
    None.

    """
    # assign lon and lat for cts_var_map
    data = wd.assign_station_meta(data, meta)

    bmp.cts_var_map(data, descr="wind/" + figmeta + "_rain-wdirn", bmap=bmap,
                    fig=fig, clabel="rainfall with "+desc+"("+unit+")",
                    vlims=vlims, log=log)


def rain_plot_fg(data, meta, bmap, fig, axesdict, facetvar='season',
                 vlims=[0.33, 3], unit='relative to mean', desc="W or NW flow",
                 figmeta="W_NW"+"_I", log=True):
    """
    Maps the observed relative rain rate occurring at stations by facetter.

    A short-cut for the special case of the `cts_var_map_fg` function from
    basemap_plots.py where the data to be mapped are the relative rain rate
    occurring for a specific facetting variable, which defaults to season.

    Parameters are equivalent to `cts_var_map_fg` except for:
    ----------
    meta : pandas DataFrame
        a list of stations, with "station" index, and 'lat' and 'lon' columns
        containing their latitude and longitude in degrees N and E,
        respectively.
    facetvar : The default is 'season'.
    vlims : The default is [0.33, 3].
    unit : The default is 'relative to mean'.
    desc : The default is "W or NW flow".
    figmeta: The default is "W_NW"+"_I".
    log : The default is True.

    Returns
    -------
    None.

    """
    # assign lon and lat for cts_var_map
    data = wd.assign_station_meta(data, meta)

    bmp.cts_var_map_fg(data, descr=figmeta + "_rain-wdirn", fig=fig,
                       axdict=axesdict, vlims=vlims, log=log, bmap=bmap,
                       clabel="rainfall with " + desc + " (" + unit + ")")


def wdirn_rain_plot(data, meta, cbarlab, unit, figmeta, bmap, fig, axesdict,
                    vlims=[0, 1], log=False, cwtname='CWT'):
    """
    Maps the proportion of rainfall occurring under each CWT at each station.

    A short-cut for the special case of the `cts_var_map_fg` function from
    basemap_plots.py where the data to be mapped are the proportion of rain
    occurring at a list of stations under a list of CWTs, or equivalent.

    Parameters are equivalent to `cts_var_map_fg` except for:
    ----------
    meta : pandas DataFrame
        a list of stations, with "station" index, and 'lat' and 'lon' columns
        containing their latitude and longitude in degrees N and E,
        respectively.
    cwtname : str
        The faceting dimension name.
    The default values of vlims are set to [0, 1]
    figmeta has the string "_by_cwt" appended to it.
    The `clabel` parameter to cts_var_map_fg is constructed from the `cbarlab`
    and `unit` parameters.
    """



    # assign lon and lat for cts_var_map
    data = wd.assign_station_meta(data, meta)

    bmp.cts_var_map_fg(data, facetvar=cwtname, bmap=bmap, fig=fig,
                       descr=figmeta + "_by_cwt", axdict=axesdict,
                       clabel=cbarlab+"("+unit+")", vlims=vlims, log=log)

# %% CWT frequency trend plots


def freq_trend_plot(data, descr, dirns=['W', 'NW'], width=150, height=100,
                    period='year', cwtname='CWT'):
    """
    Plots frequency of occurrence over time, summed over specified categories.

    By default, given annual CWT frequency of occurrence data, this function
    plots and returns as xarray DataArray the total frequency of occurrence of
    Westerly or North-Westerly synoptic flow.

    Parameters
    ----------
    data : xarray DataArray
        A 2-D frequency variable to be plotted. One dimension should be a time
        dimension (most often year), the other a category dimension, for which
        the total frequency of occurrence of the values in ``dirns`` is to be
        determined.
    descr : str
        A description of the plot to be used for the file name: must have no
        spaces.
    dirns : list or array-like of str, optional
        The values of ``cwtname`` for which the total frequency of occurrence
        is to be calculated. The default is ['W', 'NW'].
    width : float, optional
        plot width in mm. The default is 150.
    height : float, optional
        plot width in mm. The default is 100.
    period : str, optional
        Time period to be used for unit. The default is 'year'.
    cwtname : str, optional
        The name of the dimension to cumulate frequencies over.
        The default is 'CWT'.

    Returns
    -------
    data : xarray DataArray
        The plotted frequency data

    """
    fig = plt.figure(figsize=(width / 25, height / 25))
    ax = fig.add_subplot(1, 1, 1)

    # 0.25 6hrly => daily
    data = (0.25 * data.sel({cwtname: dirns})).sum(dim=cwtname)
    data.plot(ax=ax)
    ax.set(ylabel="frequency (days/" + period + ")", title='')
    figname = figdir + descr + ".png"
    fig.savefig(figname, bbox_inches='tight')
    plt.close('all')
    return data


def freq_trend_plot_fg(data, descr, facetvar='CWT', ncol=5, xrename={},
                       facet_width=120, facet_height=95, period='year'):
    """
    A facetgrid version of freq_trend_plot.

    This function is primarily useful for seasonal plots, or plots for
    multiple different CWTs/wind directions/flow regimes.

    Parameters
    ----------
    data : xarray DataArray
        A 2-D frequency variable to be plotted. One dimension should be a time
        dimension (most often year), the other a facetting dimension, with one
        value for each facet to be plotted.
    descr : str
        A description of the plot to be used for the file name: must have no
        spaces.
    facetvar : str, optional
        Name of the dimension to facet by. The default is 'CWT'.
    ncol : int, optional
        Number of columns in the facetgrid. The default is 5.
    xrename : dict, optional
        change the name of plotted time dimension. The default is {}.
        If the data are to be plotted by 'syr' (seasonally coherent year),
        but the x-axis should be labelled 'year' pass:
        ``xrename={'syr': 'year'}``
    facet_width : float, optional
        Width of each facet in mm. The default is 120.
    facet_height : float, optional
        Height of each facet in mm. The default is 95.
    period : str, optional
        Time period to be used for unit. The default is 'year'.

    Returns
    -------
    None.

    """
    fs = facet_height / 25  # facet size: height in inches
    fa = facet_width / facet_height  # aspect
    data = (0.25 * data).assign_attrs({'long_name': 'frequency',
                                       'units': 'days/'+period})

    if xrename:  # different dimension name for time axis to be used
        data = data.rename(xrename)

    # for labels
    facetvals = data[facetvar].values
    fg = data.plot(col=facetvar, col_wrap=ncol, size=fs, aspect=fa)
    # 0.25 6hrly => daily
    for i, ax in enumerate(fg.axes.flat):
        ax.set_title(facetvals[i])

    figname = figdir + descr + "_by_" + facetvar + ".png"
    plt.savefig(figname, bbox_inches='tight')
    plt.close('all')
