#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 28 14:00:26 2020
Computes climatological data for various variables

@author: stefaan
"""

import pandas as pd  # for to_dataframe()
import xarray as xr
import numpy as np

# mean length of each month
months = np.array((31, 28.25, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31))


def climatology(datavar):
    """
    Computes the climatological mean of the monthly sum of an xarray
    DataArray variable and returns three data arrays:
    the raw climatology, the climatological proportion of the annual total
    for each month for all months and for the first eleven months, given that
    the December mean proportion = 1 - sum(Jan:Nov) proportions
    Missing values are accounted for, assuming they occur at random.
    The input dataArray must have a 'time' dimension, which is a
    xarray.CFTimeIndex

    Parameters
    ----------
    datavar : xarray DataArray
        The data variable to compute the climatology of.

    Returns
    -------
    clim : xarray DataArray
        Mean monthly total of datavar
    nclim : xarray DataArray
        Mean contribution of each month to the annual total
    nclim11 : xarray DataArray
        The same as nclim for Jan-Nov only
    """

    annual = np.dot(datavar.groupby('time.month').mean(dim='time').T, months)
    clim = (datavar.groupby('time.month').mean(dim='time').T*months).T
    nclim = clim/annual
    # remove Dec; not independent as fractions add to 1
    nclim11 = nclim.where(nclim.month < 12, drop=True)
    return clim, nclim, nclim11


def norm_clim(datavar):
    """
    As for the `climatology` function, but returns only the normalised
    mean monthly totals

    Parameters
    ----------
    datavar :  xarray DataArray
        The data variable to compute the climatology of.

    Returns
    -------
    xarray DataArray
        Mean contribution of each month to the annual total

    """
    annual = np.dot(datavar.groupby('time.month').mean(dim='time').T, months)
    clim = (datavar.groupby('time.month').mean(dim='time').T*months).T
    return(clim/annual)


def climmean(datavar, by='month'):
    """
    Returns the mean of a xarray DataArray or all variables in an
    xarray Dataset, aggregated by the time period specified as `by`.
    The input DataArray or Dataset must have a 'time' dimension, which is a
    xarray.CFTimeIndex

    Parameters
    ----------
    datavar : xarray DataArray or Dataset
        The data variable to compute the climatology of or a Dataset
        with variables to compute means of.
    by : str, optional
        The time units within which the mean is calculated.
        The default is 'month'. If an empty string is supplied, the overall
        mean is returned.
        Other options include 'season', 'year' and 'dayofyear' or any other
        time period attribute of a xarray.CFTimeIndex object


    Returns
    -------
    Same as datavar
        Contains the mean for each unique value of the `by` parameter
    """

    if len(by) > 0:
        return datavar.groupby('time.'+by).mean(dim='time')
    else:
        return datavar.mean(dim='time')


def climsd(datavar, by='month'):
    """
    Returns the standard deviation of an xarray DataArray or all variables
    in an xarray Dataset, aggregated by the time period specified as `by`.
    The input dataArray must have a 'time' dimension, which is a
    xarray.CFTimeIndex

    Parameters
    ----------
    datavar :  xarray DataArray or Dataset
        The data variable to compute the climatology of or a Dataset
        with variables to compute means of.
    by : str, optional
        The time units within which the standard deviation is calculated.
        The default is 'month'. If an empty string is supplied, the overall
        mean is returned.
        Other options include 'season', 'year' and 'dayofyear' or any other
        time period attribute of a xarray.CFTimeIndex object

    Returns
    -------
    Same as datavar
        Contains the standard deviation for each unique value of the `by`
        parameter

    """
    if len(by) > 0:
        return datavar.groupby('time.'+by).sd(dim='time')
    else:
        return datavar.mean(dim='time')


def ann_climatology(datavar):
    """
    As for climatology, but provides mean annual rather than mean monthly sums.

    Parameters
    ----------
    datavar : xarray DataArray
        The data variable to compute the climatology of.
    Returns
    -------
    xarray DataArray
        The mean annual sum
    """

    return climatology(datavar)[0].sum(dim='month')


def rd_clim(datavar, thres=0.2):
    """
    Computes the monthly rain day climatology given a xarray DataArray
    containing daily rainfall data. Days receiving at least `thres` rainfall
    are considered as rain days. Returns the same three outputs as the
    `climatology function`. This function does not work correctly
    in the presence of missing values. The input dataArray must have a
    'time' dimension, which is a xarray.CFTimeIndex

    Parameters
    ----------
    datavar : xarray DataArray
        Daily rainfall DataArray
    thres : float, optional
        The minimum rainfall amount that is The default is 0.2.

    Returns
    -------
    three xarray DataArrays
        See `climatology` for details

    """
    return climatology(datavar >= thres)


def rd_annclim(datavar, thres=0.2):
    """
    Computes the annual rain day climatology given an xarray DataArray
    containing daily rainfall data. Days receiving at least `thres` rainfall
    are considered as rain days. This function does not work correctly
    in the presence of missing values. The input dataArray must have a
    'time' dimension, which is a xarray.CFTimeIndex

    Parameters
    ----------
    see `rd_clim`

    Returns
    -------
    see `ann_climatology`
    """
    return rd_clim(datavar, thres=thres)[0].sum(dim='month')


def rd_climna(datavar, thres=0.2):
    """As with rd_clim, but works in the presence of missing data."""

    return(climatology(np.greater(datavar, thres,
                                  out=xr.full_like(datavar, np.nan).values,
                                  where=datavar.notnull())))


def rd_annclimna(datavar, thres=0.2):
    """As with rd_annclim, but works in the presence of missing data."""
    return rd_climna(datavar, thres=thres)[0].sum(dim='month')


def climatology_station(dataset, station, var='pr', stationvar='name',
                        idvar='id', stationdim='feature'):
    """
    For a station xarray Dataset, where stations are included along a
    `stationdim` dimension, for a selected station, identified by passing
    either the station name (stored in `stationvar`) or id (stored in `idvar`)
    to the `station` parameter, returns the mean monthly totals for that
    station, for the variable `var`.


    Parameters
    ----------
    dataset : xarray Dataset
        The Dataset containing the data to aggregate
    station : str
        Value of the `idvar` or `stationvar` variable that identifies the
        station of interest.
    var : , optional
        DESCRIPTION. The default is 'pr'.
    stationvar : TYPE, optional
        DESCRIPTION. The default is 'name'.
    idvar : TYPE, optional
        DESCRIPTION. The default is 'id'.
    stationdim : TYPE, optional
        DESCRIPTION. The default is 'feature'.

    Returns
    -------
    xarray DataArray
        The climatological monthly mean totals for the station in question.

    """
    station = dataset.sel({stationdim: ((dataset[stationvar] == station) |
                                        (dataset[id] == station))}, drop=True)
    return station[var].groupby('time.month').mean(dim='time').T*months


def norm_climatology_station(dataset, station, var='pr', stationvar='name',
                             idvar='id', stationdim='feature'):
    """ As with `climatology_station`, but returns the climatologies
    normalised by the annual total (i.e. the climatological proportion)
    occurring during a given month"""
    station = dataset.sel({stationdim: ((dataset[stationvar] == station) |
                                        (dataset[idvar] == station))},
                          drop=True)

    annual = np.dot(station[var].groupby('time.month').mean(dim='time').T,
                    months)
    return station[var].groupby('time.month').mean(dim='time').T*months/annual


def AMJJAS(month):
    """
    Determines whether a given month, or array of months, fall within the
    extended Austral winter season (April-September)

    Parameters
    ----------
    month : int or array-like of ints
        The month or months to test

    Returns
    -------
    bool or array-like of bools
        Whether each month falls in AMJJAS

    """
    return (month >= 4) & (month <= 9)


def inseas(month, seasmonths):
    """tests whether `month` is in the season defined by
    the set of months `seasmonths`

    Parameters
    ----------
    month : int or array-like of ints
        The month or months to test

    seasmonth : array or list of ints
        containing the months that make up the season. Should all be between
        1 and 12

    Returns
    -------
    bool or array-like of bools
        Whether each month falls in the user-defined season

    """
    return month.isin(seasmonths)


def assign_hyr(data):
    """
    Takes an xarray Dataset and returns the dataset with all variables plus
    a new 'hyr' variable, for the hydrological year, starting in October
    of the preceding calendar year. Only data in complete hydrological years
    are returned. The input Dataset must have a 'time' dimension, which is a
    xarray.CFTimeIndex

    Parameters
    ----------
    data : xarray Dataset
        The dataset to add the hyr variable to

    Returns
    -------
    Dataset
        The original input with the hyr variable added with data from
        incomplete hydrological years discarded.

    """
    data = data.assign(hyr=(('time'),
                            data['time.year'].where(data['time.month'] < 10,
                                                    data['time.year'] + 1)))
    i0 = list(data['time.month'].values).index(10)  # first index in October
    # first index of final hyr data, which will be discarded;
    # assumes data don't end in Sept
    lastyr = data['hyr'].max()
    iN = list(data['hyr'].values).index(lastyr)
    return data.isel(time=slice(i0, iN))


def assign_syr(data, mon0=12):
    """
    Takes an xarray Dataset and returns the dataset with all variables plus
    a new 'syr' variable, for the seasonal year, starting in `mon0`
    of the preceding calendar year. It is intended to be used to ensure
    contiguous seasonal blocks are used for seasonal aggregation.
    Only data in complete hydrological years are returned. The input
    Dataset must have a 'time' dimension, which is a xarray.CFTimeIndex

    Parameters
    ----------
    data : xarray Dataset
        The dataset to add the syr variable to
    mon0 : int, optional
        The month that syr starts in. The default is 12. Must be in [1, 12].

    Returns
    -------
    Dataset
        The original input with the syr variable added with data from
        incomplete syrs discarded.

    """
    data = data.assign(syr=(('time'),
                            data['time.year'].where(data['time.month'] < mon0,
                                                    data['time.year'] + 1
                                                    ).data))

    i0 = list(data['time.month'].values).index(mon0)  # first index in mon0

    # throw out all data before start of first syr

    # first index of final syr data, which will be discarded
    # (assuming data are for complete calendar years)
    lastyr = data['syr'].max()
    iN = list(data['syr'].values).index(lastyr)
    return data.isel(time=slice(i0, iN))


def ia_tot(data, yrvar='time.year'):
    """
    Interannual variable totals, transformed to a fixed 365.25-day year,
    accounting for random missing values. The input data must have a
    'time' dimension, which is a xarray.CFTimeIndex

    Parameters
    ----------
    data : xarray DataArray or Dataset
        The data to be aggregated
    yrvar : str, optional
        The variable to group data into years by. The default is 'time.year'.

    Returns
    -------
    same as `data` input
        The annually aggregated data
    """
    return data.groupby(yrvar).mean(dim='time')*365.25


def ia_ew_tot(data, yrvar='time.year'):
    """ As with ia_tot, but considering only data for the austral extended
    winter period (AMJJAS)"""
    return (data.sel(time=AMJJAS(data['time.month'])).groupby(yrvar)
            ).mean(dim='time')*183


def ia(data, yrvar='hyr', var='pr'):
    """
    Computes annual anomaly data for a variable in an xarray Dataset.
    By default data are aggregated by hydrological rathern than calendar year.
    The Dataset `data` should have a xarray.CFTimeIndex 'time' dimension.

    Parameters
    ----------
    data : Dataset
       The dataset containing the variable to be aggregated
    yrvar : str, optional
        The variable by which aggregation is to be performed.
        The default is 'hyr'. For calendar year use 'time.year'
    var : str, optional
        The variable in the `data` to be aggregated. The default is 'pr'.

    Returns
    -------
    annual : xarray DataArray
        The annual anomaly data variable
    """
    annual = data.groupby(yrvar).mean(dim='time')[var]*365.25
    # groupby requires this if 'time.year' is used
    yrvar = yrvar.lstrip('time.')
    annual = (annual - annual.mean(dim=yrvar))  # anomalies
    return annual


def std_ia(data, groupvar='time.year', yrvar='year'):
    """As with `ia` but returning standardised anomalies (anomalies divided
    by corresponding standard deviations)"""
    annual = data.groupby(groupvar).mean(dim='time')*365.25
    # std. anomalies
    annual = (annual-annual.mean(dim=yrvar))/annual.std(dim=yrvar)
    return annual


def std_ia_ew(data, groupvar='time.year', drop2=False):
    """
    As with std_ia, but considering only data observed during the extended
    Austral winter semester (AMJJAS).
    The DataArray `data` should have a xarray.CFTimeIndex 'time' dimension.

    Parameters
    ----------
    data : xarray DataArray
        Variable to manipulated
    groupvar : str, optional
         The time variable to use for annual aggregation.
         The default is 'time.year'.
    drop2 : bool, optional
        Whether to remove data made redundant by standardisation of anomalies.
        The default is False.

    Returns
    -------
    xarray DataArray
        Containing the standardised anomaly data
    """
    data = data.sel(time=AMJJAS(data['time.month']))
    annual = data.groupby(groupvar).mean(dim='time') * 183

    # if the grouping variable is of the form 'time.x' the 'time.' part
    # will be removed from the dimension name of the result
    yrvar = groupvar.lstrip('time.')
    # std. anomalies
    annual = (annual-annual.mean(dim=yrvar))/annual.std(dim=yrvar)
    if drop2:  # retain only the sign for yr1
        annual = annual.where(annual[yrvar] != 1,
                              other=np.sign(annual.sel({yrvar: 1})))
        return annual.where(annual[yrvar] != 0)  # drop the first year
    else:
        return annual


def ia_ew(data, yrvar='hyr', var='pr'):
    """As with `ia`, but considering only the extended Austral winter
    semester (AMJJAS) """
    winter = data.sel(time=AMJJAS(data['time.month']))
    winter = winter.groupby(yrvar).mean(dim='time')[var]*183
    yrvar = yrvar.lstrip('time.')
    winter = (winter - winter.mean(dim=yrvar))  # anomalies
    return winter


def ia_seas(data, seas, yrvar='syr', var='pr'):
    """
    As for ia_ew, but returns anomalies computed for one of the standard
    3-month meteoroloqgical seasons defined by the input parameter `seas`.

    Parameters
    ----------
    data : xarray Dataset
        The Dataset containing the variable to be aggregated
    seas : str
        One of 'DJF', 'MAM', 'JJA' or 'SON'
    yrvar : str, optional
        The annual variable to collect seasons by. The default is 'syr'.
    var : str, optional
        The variable to aggregate. The default is 'pr'.

    Returns
    -------
    data : xarray DataArray
        The seasonal anomalies of the variable considered.
    """
    nyrs = len(list(set(data['time.year'].values)))
    # data for chosen season only
    data = data.sel(time=data['time.season'] == seas)
    nt = data.dims['time'] / nyrs  # avg days/seas
    data = data.groupby(yrvar).mean(dim='time')[var] * nt
    data = (data - data.mean(dim=yrvar))  # anomalies
    return data


def prop_valid(data, yrvar='hyr', timevar='time.season',
               timeval='JJA', var='pr', thres=-1e4):
    """
    The fraction of data in the time period defined by `timevar` == `timeval`
    in the variable `var` of the Dataset `data` which are not missing during
    each year within the data period.
    The DataArray `data` should have a xarray.CFTimeIndex 'time' dimension.
    Values less than `thres` are assumed to represent missing values

    Parameters
    ----------
    data : xarray Dataset
        Dataset containing the variable to be considered
    yrvar : str, optional
        The annual aggregation variable. The default is 'hyr'.
    timevar : TYPE, optional
        The time variable to use for assessment. The default is 'time.season'.
    timeval : TYPE, optional
        The time period to consider. Must be a valid value of `timevar`.
        The default is 'JJA'.
    var : str, optional
        The variable to consider. The default is 'pr'.
    thres: float
        Values below this threshold are assumed to represent invalid data

    Returns
    -------
    data : xarray DataArray
        The proportion (fraction) of the considered data that are not missing
        during each year

    """
    data = data.sel(time=data[timevar] == timeval)
    data['valid'] = data[var] > thres
    # equivalent to prop valid
    data = data.groupby(yrvar).mean(dim='time')['valid']
    return data


def ew_vs_es(clim, var="pr", latvar="latitude", lonvar="longitude"):
    """
    Taking in a climatological xarray Dataset with dimension 'month',
    returns a pandas dataframe containing the ratio of mean extended Austral
    winter (AMJJAS) sum to the extended Austral summer (ONDJFM) mean sum
    of the variable of interest, at each location.

    Parameters
    ----------
    clim : xarray Dataset
        Contains monthly climatological data for `var`.
    var : str, optional
        Climatological variable of interest. The default is "pr".
    latvar : str, optional
        The name of the latitude dimension in the Dataset.
        The default is "latitude".
    lonvar : str, optional
        The name of the latitude dimension in the Dataset.
        The default is "longitude".

    Returns
    -------
    pandas dataframe
        A dataframe with the ratio of interest in the variable "WvS" and the
        latitude variable renamed to "lat" and the longitude to "lon".
    """
    ew_clim = clim.sel(month=AMJJAS(clim['month'])).sum(dim='month')
    es_clim = clim.sum(dim='month') - ew_clim  # annual - winter = summer
    ew_v_es = (ew_clim/es_clim).to_dataframe().reset_index()
    return ew_v_es.rename(columns={latvar: "lat", lonvar: "lon",
                                   var: "WvS"})


def ew_vs_yr(clim, var="pr", latvar="latitude", lonvar="longitude",
             perc=True):
    """
    Taking in a climatological xarray Dataset with dimension 'month',
    returns a pandas dataframe containing the ratio of mean extended Austral
    winter (AMJJAS) sum to the annual mean sum of the variable of interest,
    at each location.

    Parameters
    ----------
    clim : xarray Dataset
        Contains monthly climatological data for `var`.
    var : str, optional
        Climatological variable of interest. The default is "pr".
    latvar : str, optional
        The name of the latitude dimension in the Dataset.
        The default is "latitude".
    lonvar : str, optional
        The name of the latitude dimension in the Dataset.
        The default is "longitude".
    perc : bool, optional
        Whether the resultant ratio as a percentage

    Returns
    -------
    pandas dataframe
        A dataframe with the ratio of interest in the variable "WvYr" and the
        latitude variable renamed to "lat" and the longitude to "lon".
    """
    ew_clim = clim.sel(month=AMJJAS(clim['month'])).sum(dim='month')
    yr_clim = clim.sum(dim='month')  # annual
    ew_v_yr = (ew_clim/yr_clim).to_dataframe().reset_index()
    ew_v_yr[var] = (1 + 99 * perc) * ew_v_yr[var]  # convert to % if req.
    return ew_v_yr.rename(columns={latvar: "lat", lonvar: "lon",
                                   var: "WvYr"})


def perc_seas_monclim(xrda, mon1=4, mon2=9):
    """
    Calculate the mean percentage of a variable occurring during the season
    between mon1 and mon2. Defaults to AMJJAS as season. The data must be
    monthly and must have 'month' as a dimension. First calculates the monthly
    mean values, so that mean values of each month carry equal weight
    rather than being weighted by their length.

    Parameters
    ----------
    xrda : xarray DataArray
        Variable of interest
    mon1 : int, between 1 and 12, optional
        The starting month of the season to consider. The default is 4.
    mon2 : int, between 1 and 12, optional
        The final month of the season to consider. The default is 9.

    Returns
    -------
    xarray DataArray
        Array containing the climatological percentage of rainfall occurring
        during the specified season.
    """
    monclm = climmean(xrda)
    return ((monclm.sel(month=inseas(monclm['month'], np.arange(mon1,
                                                                mon2 + 1))
                        ).sum(dim='month')) * 100 / (monclm.sum(dim='month')))


def perc_seas(xrda, mon1=4, mon2=9):
    """
    Calculate the mean percentage of a variable occurring during the season
    between mon1 and mon2. Defaults to AMJJAS as season. The data must be
    monthly and must have 'month' as a dimension.

    Parameters
    ----------
    xrda : xarray DataArray
        Variable of interest
    mon1 : int, between 1 and 12, optional
        The starting month of the season to consider. The default is 4.
    mon2 : int, between 1 and 12, optional
        The final month of the season to consider. The default is 9.

    Returns
    -------
    xarray DataArray
        Array containing the climatological percentage of rainfall occurring
        during the specified season.
    """
    seasmu = xrda.sel(time=inseas(xrda['time.month'],
                                  np.arange(mon1, mon2 + 1))
                      ).mean(dim='time')
    yrmu = xrda.mean(dim='time')
    return 100 * (seasmu/yrmu)*183/365.25
